# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from pathlib import Path

from drgn import (
    ExtraModule,
    MainModule,
    ModuleFileStatus,
    Program,
    RelocatableModule,
    SharedLibraryModule,
    VdsoModule,
)
from tests import IntWrapper, TestCase


class TestModule(TestCase):
    def _test_module_init_common(self, module):
        self.assertIsNone(module.address_ranges)
        self.assertIsNone(module.address_range)
        self.assertIsNone(module.build_id)
        self.assertEqual(module.loaded_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.loaded_file_path)
        self.assertIsNone(module.loaded_file_bias)
        self.assertEqual(module.debug_file_status, ModuleFileStatus.WANT)
        self.assertIsNone(module.debug_file_path)
        self.assertIsNone(module.debug_file_bias)
        self.assertIsNone(module.supplementary_debug_file_kind)
        self.assertIsNone(module.supplementary_debug_file_path)

    def test_main_module(self):
        prog = Program()

        self.assertRaises(LookupError, prog.main_module)
        self.assertRaises(LookupError, prog.main_module, "/foo/bar")

        module = prog.main_module("/foo/bar", create=True)
        self.assertIsInstance(module, MainModule)

        self.assertEqual(prog.main_module(), module)
        self.assertEqual(prog.main_module(create=False), module)
        self.assertEqual(prog.main_module("/foo/bar"), module)
        self.assertEqual(prog.main_module(b"/foo/bar"), module)
        self.assertEqual(prog.main_module(Path("/foo/bar")), module)
        self.assertEqual(prog.main_module("/foo/bar", create=True), module)

        self.assertRaises(LookupError, prog.main_module, "/foo/baz")
        self.assertRaises(LookupError, prog.main_module, "/foo/baz", create=True)

        self.assertIs(module.prog, prog)
        self.assertEqual(module.name, "/foo/bar")
        self._test_module_init_common(module)

    def test_main_module_invalid(self):
        prog = Program()
        self.assertRaises(TypeError, prog.main_module, None)
        self.assertRaises(TypeError, prog.main_module, create=True)
        self.assertRaises(TypeError, prog.main_module, "/foo/bar", True)

    def test_shared_library_module(self):
        prog = Program()

        self.assertRaises(
            LookupError, prog.shared_library_module, "/foo/bar", 0x10000000
        )

        module = prog.shared_library_module("/foo/bar", 0x10000000, create=True)
        self.assertIsInstance(module, SharedLibraryModule)

        self.assertEqual(prog.shared_library_module("/foo/bar", 0x10000000), module)
        self.assertEqual(prog.shared_library_module(b"/foo/bar", 0x10000000), module)
        self.assertEqual(
            prog.shared_library_module(Path("/foo/bar"), IntWrapper(0x10000000)), module
        )
        self.assertEqual(
            prog.shared_library_module("/foo/bar", 0x10000000, create=True), module
        )

        self.assertRaises(
            LookupError, prog.shared_library_module, "/foo/bar", 0x20000000
        )
        self.assertRaises(
            LookupError, prog.shared_library_module, "/foo/baz", 0x10000000
        )

        self.assertNotEqual(
            prog.shared_library_module("/foo/bar", 0x20000000, create=True), module
        )
        self.assertNotEqual(
            prog.shared_library_module("/foo/baz", 0x10000000, create=True), module
        )
        self.assertNotEqual(
            prog.vdso_module("/foo/bar", 0x10000000, create=True), module
        )

        self.assertIs(module.prog, prog)
        self.assertEqual(module.name, "/foo/bar")
        self.assertEqual(module.dynamic_address, 0x10000000)
        self._test_module_init_common(module)

    def test_shared_library_module_invalid(self):
        prog = Program()
        self.assertRaises(TypeError, prog.shared_library_module)
        self.assertRaises(TypeError, prog.shared_library_module, "/foo/bar")
        self.assertRaises(TypeError, prog.shared_library_module, "/foo/bar", None)
        self.assertRaises(TypeError, prog.shared_library_module, None, 0)
        self.assertRaises(
            TypeError, prog.shared_library_module, "/foo/bar", 0x10000000, True
        )

    def test_vdso_module(self):
        prog = Program()

        self.assertRaises(LookupError, prog.vdso_module, "/foo/bar", 0x10000000)

        module = prog.vdso_module("/foo/bar", 0x10000000, create=True)
        self.assertIsInstance(module, VdsoModule)

        self.assertEqual(prog.vdso_module("/foo/bar", 0x10000000), module)
        self.assertEqual(prog.vdso_module(b"/foo/bar", 0x10000000), module)
        self.assertEqual(
            prog.vdso_module(Path("/foo/bar"), IntWrapper(0x10000000)), module
        )
        self.assertEqual(prog.vdso_module("/foo/bar", 0x10000000, create=True), module)

        self.assertRaises(LookupError, prog.vdso_module, "/foo/bar", 0x20000000)
        self.assertRaises(LookupError, prog.vdso_module, "/foo/baz", 0x10000000)

        self.assertNotEqual(
            prog.vdso_module("/foo/bar", 0x20000000, create=True), module
        )
        self.assertNotEqual(
            prog.vdso_module("/foo/baz", 0x10000000, create=True), module
        )
        self.assertNotEqual(
            prog.shared_library_module("/foo/bar", 0x10000000, create=True), module
        )

        self.assertIs(module.prog, prog)
        self.assertEqual(module.name, "/foo/bar")
        self.assertEqual(module.dynamic_address, 0x10000000)
        self._test_module_init_common(module)

    def test_vdso_module_invalid(self):
        prog = Program()
        self.assertRaises(TypeError, prog.vdso_module)
        self.assertRaises(TypeError, prog.vdso_module, "/foo/bar")
        self.assertRaises(TypeError, prog.vdso_module, "/foo/bar", None)
        self.assertRaises(TypeError, prog.vdso_module, None, 0)
        self.assertRaises(TypeError, prog.vdso_module, "/foo/bar", 0x10000000, True)

    def test_relocatable_module(self):
        prog = Program()

        self.assertRaises(LookupError, prog.relocatable_module, "/foo/bar", 0x10000000)

        module = prog.relocatable_module("/foo/bar", 0x10000000, create=True)
        self.assertIsInstance(module, RelocatableModule)

        self.assertEqual(prog.relocatable_module("/foo/bar", 0x10000000), module)
        self.assertEqual(prog.relocatable_module(b"/foo/bar", 0x10000000), module)
        self.assertEqual(
            prog.relocatable_module(Path("/foo/bar"), IntWrapper(0x10000000)), module
        )
        self.assertEqual(
            prog.relocatable_module("/foo/bar", 0x10000000, create=True), module
        )

        self.assertRaises(LookupError, prog.relocatable_module, "/foo/bar", 0x20000000)
        self.assertRaises(LookupError, prog.relocatable_module, "/foo/baz", 0x10000000)

        self.assertNotEqual(
            prog.relocatable_module("/foo/bar", 0x20000000, create=True), module
        )
        self.assertNotEqual(
            prog.relocatable_module("/foo/baz", 0x10000000, create=True), module
        )
        self.assertNotEqual(
            prog.shared_library_module("/foo/bar", 0x10000000, create=True), module
        )

        self.assertIs(module.prog, prog)
        self.assertEqual(module.name, "/foo/bar")
        self.assertEqual(module.address, 0x10000000)
        self._test_module_init_common(module)

    def test_section_addresses(self):
        prog = Program()
        module = prog.relocatable_module("/foo/bar", 0x10000000, create=True)

        self.assertNotIn(".text", module.section_addresses)
        self.assertNotIn(1, module.section_addresses)

        with self.assertRaises(KeyError):
            module.section_addresses[".text"]
        with self.assertRaises(KeyError):
            module.section_addresses[1]

        with self.assertRaises(KeyError):
            del module.section_addresses[".text"]
        with self.assertRaises(KeyError):
            del module.section_addresses[1]

        module.section_addresses[".text"] = 0x10000000
        self.assertIn(".text", module.section_addresses)
        self.assertEqual(module.section_addresses[".text"], 0x10000000)

        self.assertEqual(len(module.section_addresses), 1)
        self.assertCountEqual(list(module.section_addresses), [".text"])
        self.assertCountEqual(list(module.section_addresses.keys()), [".text"])
        self.assertCountEqual(list(module.section_addresses.values()), [0x10000000])
        self.assertCountEqual(
            list(module.section_addresses.items()), [(".text", 0x10000000)]
        )

        module.section_addresses[".data"] = 0x10001000

        self.assertEqual(len(module.section_addresses), 2)
        self.assertCountEqual(list(module.section_addresses), [".text", ".data"])
        self.assertCountEqual(list(module.section_addresses.keys()), [".text", ".data"])
        self.assertCountEqual(
            list(module.section_addresses.values()), [0x10000000, 0x10001000]
        )
        self.assertCountEqual(
            list(module.section_addresses.items()),
            [(".text", 0x10000000), (".data", 0x10001000)],
        )

        del module.section_addresses[".data"]

        self.assertEqual(len(module.section_addresses), 1)
        self.assertCountEqual(list(module.section_addresses), [".text"])
        self.assertCountEqual(list(module.section_addresses.keys()), [".text"])
        self.assertCountEqual(list(module.section_addresses.values()), [0x10000000])
        self.assertCountEqual(
            list(module.section_addresses.items()), [(".text", 0x10000000)]
        )

    def test_relocatable_module_invalid(self):
        prog = Program()
        self.assertRaises(TypeError, prog.relocatable_module)
        self.assertRaises(TypeError, prog.relocatable_module, "/foo/bar")
        self.assertRaises(TypeError, prog.relocatable_module, "/foo/bar", None)
        self.assertRaises(TypeError, prog.relocatable_module, None, 0)
        self.assertRaises(
            TypeError, prog.relocatable_module, "/foo/bar", 0x10000000, True
        )

    def test_extra_module(self):
        prog = Program()

        self.assertRaises(LookupError, prog.extra_module, "/foo/bar", 1234)

        module = prog.extra_module("/foo/bar", 1234, create=True)
        self.assertIsInstance(module, ExtraModule)

        self.assertEqual(prog.extra_module("/foo/bar", 1234), module)
        self.assertEqual(prog.extra_module(b"/foo/bar", 1234), module)
        self.assertEqual(prog.extra_module(Path("/foo/bar"), IntWrapper(1234)), module)
        self.assertEqual(prog.extra_module("/foo/bar", 1234, create=True), module)

        self.assertRaises(LookupError, prog.extra_module, "/foo/bar", 5678)
        self.assertRaises(LookupError, prog.extra_module, "/foo/baz", 1234)

        self.assertNotEqual(prog.extra_module("/foo/bar", 5678, create=True), module)
        self.assertNotEqual(prog.extra_module("/foo/baz", 1234, create=True), module)
        self.assertNotEqual(
            prog.shared_library_module("/foo/bar", 1234, create=True), module
        )
        self.assertEqual(prog.extra_module("/foo/bar", create=True).id, 0)

        self.assertIs(module.prog, prog)
        self.assertEqual(module.name, "/foo/bar")
        self.assertEqual(module.id, 1234)
        self._test_module_init_common(module)

    def test_extra_module_invalid(self):
        prog = Program()
        self.assertRaises(TypeError, prog.extra_module)
        self.assertRaises(TypeError, prog.extra_module, "/foo/bar", None)
        self.assertRaises(TypeError, prog.extra_module, None, 0)
        self.assertRaises(TypeError, prog.extra_module, "/foo/bar", 1234, True)

    def test_address_range(self):
        module = Program().extra_module("/foo/bar", create=True)

        module.address_range = (0x10000000, 0x10010000)
        self.assertEqual(module.address_range, (0x10000000, 0x10010000))

        module.address_range = (0x20000000, 0x20020000)
        self.assertEqual(module.address_range, (0x20000000, 0x20020000))

        module.address_range = None
        self.assertIsNone(module.address_range)

        module.address_range = None
        self.assertIsNone(module.address_range)

    def test_address_range_empty(self):
        module = Program().extra_module("/foo/bar", create=True)

        module.address_range = (0, 0)
        self.assertEqual(module.address_range, (0, 0))

    def test_address_range_type_error(self):
        module = Program().extra_module("/foo/bar", create=True)

        with self.assertRaises(TypeError):
            module.address_range = 1

        with self.assertRaises(TypeError):
            module.address_range = (1,)

        with self.assertRaises(TypeError):
            module.address_range = (1, 2, 3)

        with self.assertRaises(TypeError):
            module.address_range = ("foo", 1)

        with self.assertRaises(TypeError):
            module.address_range = (1, "bar")

    def test_address_range_invalid(self):
        module = Program().extra_module("/foo/bar", create=True)

        with self.assertRaisesRegex(ValueError, "invalid module address range"):
            module.address_range = (0x10010000, 0x10000000)

        with self.assertRaisesRegex(ValueError, "invalid module address range"):
            module.address_range = (1, 1)

        with self.assertRaisesRegex(ValueError, "invalid module address range"):
            module.address_range = (2**64 - 1, 1)

        with self.assertRaisesRegex(ValueError, "invalid module address range"):
            module.address_range = (2**64 - 1, 2**64 - 1)

    def test_address_range_del(self):
        module = Program().extra_module("/foo/bar", create=True)
        with self.assertRaises(AttributeError):
            del module.address_range

    def test_address_ranges_single(self):
        module = Program().extra_module("/foo/bar", create=True)

        module.address_ranges = [(0x10000000, 0x10010000)]
        self.assertEqual(module.address_range, (0x10000000, 0x10010000))
        self.assertCountEqual(module.address_ranges, [(0x10000000, 0x10010000)])

        module.address_range = (0x20000000, 0x20010000)
        self.assertCountEqual(module.address_ranges, [(0x20000000, 0x20010000)])

    def test_address_ranges_multiple(self):
        module = Program().extra_module("/foo/bar", create=True)

        module.address_ranges = [
            (0x10000000, 0x10010000),
            (0x20000000, 0x20010000),
        ]
        with self.assertRaisesRegex(ValueError, "module has multiple address ranges"):
            module.address_range
        self.assertCountEqual(
            module.address_ranges, [(0x10000000, 0x10010000), (0x20000000, 0x20010000)]
        )

    def test_address_ranges_empty(self):
        module = Program().extra_module("/foo/bar", create=True)

        module.address_ranges = ()
        self.assertEqual(module.address_range, (0, 0))
        self.assertCountEqual(module.address_ranges, ())

        module.address_range = (0, 0)
        self.assertCountEqual(module.address_ranges, ())

    def test_address_ranges_type_error(self):
        module = Program().extra_module("/foo/bar", create=True)

        with self.assertRaises(TypeError):
            module.address_ranges = 1

        with self.assertRaises(TypeError):
            module.address_ranges = (1,)

        with self.assertRaises(TypeError):
            module.address_ranges = ((1,),)

        with self.assertRaises(TypeError):
            module.address_ranges = ((1, 2, 3),)

        with self.assertRaises(TypeError):
            module.address_ranges = (("foo", 1),)

        with self.assertRaises(TypeError):
            module.address_ranges = ((1, "bar"),)

    def test_address_ranges_invalid(self):
        module = Program().extra_module("/foo/bar", create=True)

        with self.assertRaisesRegex(ValueError, "invalid module address range"):
            module.address_ranges = ((0x10010000, 0x10000000),)

        with self.assertRaisesRegex(ValueError, "invalid module address range"):
            module.address_ranges = ((1, 1),)

        with self.assertRaisesRegex(ValueError, "invalid module address range"):
            module.address_ranges = ((0, 0),)

    def test_address_range_overlap(self):
        prog = Program()
        prog.extra_module("/foo/bar", create=True).address_range = (
            0x10000000,
            0x10010000,
        )
        with self.assertRaisesRegex(ValueError, "overlaps /foo/bar"):
            prog.extra_module("/foo/baz", create=True).address_range = (
                0x10001000,
                0x10010000,
            )

    def test_address_ranges_overlap(self):
        prog = Program()
        prog.extra_module("/foo/bar", create=True).address_range = (
            0x10000000,
            0x10010000,
        )
        with self.assertRaisesRegex(ValueError, "overlaps /foo/bar"):
            prog.extra_module("/foo/baz", create=True).address_ranges = (
                (0x1000000, 0x10000000),
                (0x10001000, 0x10010000),
            )

    def test_address_ranges_self_overlap(self):
        module = Program().extra_module("/foo/bar", create=True)
        with self.assertRaisesRegex(ValueError, "overlaps"):
            module.address_ranges = (
                (0x10000000, 0x10010000),
                (0x10001000, 0x10020000),
            )

    def test_address_range_expand(self):
        module = Program().extra_module("/foo/bar", create=True)
        module.address_range = (0x10000000, 0x10010000)
        module.address_range = (0x1000000, 0x10020000)
        self.assertEqual(module.address_range, (0x1000000, 0x10020000))

    def test_build_id(self):
        module = Program().extra_module("/foo/bar", create=True)

        module.build_id = b"\x01\x23\x45\x67\x89\xab\xcd\xef"
        self.assertEqual(module.build_id, b"\x01\x23\x45\x67\x89\xab\xcd\xef")

        module.build_id = b"\xfe\xdc\xba\x98\x76\x54\x32\x10"
        self.assertEqual(module.build_id, b"\xfe\xdc\xba\x98\x76\x54\x32\x10")

        module.build_id = None
        self.assertIsNone(module.build_id)

        module.build_id = None
        self.assertIsNone(module.build_id)

    def test_build_id_type_error(self):
        module = Program().extra_module("/foo/bar", create=True)
        with self.assertRaises(TypeError):
            module.build_id = "abcd"

    def test_build_id_invalid_empty(self):
        module = Program().extra_module("/foo/bar", create=True)
        with self.assertRaisesRegex(ValueError, "build ID cannot be empty"):
            module.build_id = b""

    def test_build_id_del(self):
        module = Program().extra_module("/foo/bar", create=True)
        with self.assertRaises(AttributeError):
            del module.build_id

    def test_find_by_name(self):
        prog = Program()
        self.assertRaises(LookupError, prog.module, "foo")

        module1 = prog.extra_module("foo", create=True)
        self.assertEqual(prog.module("foo"), module1)

        module2 = prog.main_module("foo", create=True)
        self.assertIn(prog.module("foo"), (module1, module2))

        self.assertRaises(LookupError, prog.module, "bar")

    def test_find_by_address(self):
        prog = Program()
        module1 = prog.extra_module("/foo/bar", create=True)
        module1.address_range = (0x10000000, 0x10010000)
        module2 = prog.extra_module("/asdf/jkl", create=True)
        module2.address_range = (0x20000000, 0x20020000)

        self.assertRaises(LookupError, prog.module, 0x0FFFFFFF)
        self.assertEqual(prog.module(0x10000000), module1)
        self.assertEqual(prog.module(0x10000001), module1)
        self.assertEqual(prog.module(0x1000FFFF), module1)
        self.assertRaises(LookupError, prog.module, 0x10010000)

        self.assertRaises(LookupError, prog.module, 0x1FFFFFFF)
        self.assertEqual(prog.module(0x20000000), module2)
        self.assertEqual(prog.module(0x20000001), module2)
        self.assertEqual(prog.module(0x2001FFFF), module2)
        self.assertRaises(LookupError, prog.module, 0x20020000)

    # Test all of the state transitions that we can without setting a file.
    def _test_file_status(self, which):
        module = Program().extra_module("/foo/bar", create=True)

        status_attr = which + "_file_status"
        wants_file = getattr(module, f"wants_{which}_file")

        self.assertRaises(TypeError, setattr, module, status_attr, 1)

        setattr(module, status_attr, ModuleFileStatus.WANT)
        self.assertEqual(getattr(module, status_attr), ModuleFileStatus.WANT)
        self.assertEqual(wants_file(), True)
        for status in set(ModuleFileStatus) - {
            ModuleFileStatus.WANT,
            ModuleFileStatus.DONT_WANT,
            ModuleFileStatus.DONT_NEED,
        }:
            with self.subTest(from_=ModuleFileStatus.WANT, to=status):
                self.assertRaises(ValueError, setattr, module, status_attr, status)
                self.assertEqual(getattr(module, status_attr), ModuleFileStatus.WANT)

        setattr(module, status_attr, ModuleFileStatus.DONT_WANT)
        self.assertEqual(getattr(module, status_attr), ModuleFileStatus.DONT_WANT)
        self.assertEqual(wants_file(), False)
        for status in set(ModuleFileStatus) - {
            ModuleFileStatus.WANT,
            ModuleFileStatus.DONT_WANT,
            ModuleFileStatus.DONT_NEED,
        }:
            with self.subTest(from_=ModuleFileStatus.DONT_WANT, to=status):
                self.assertRaises(ValueError, setattr, module, status_attr, status)
                self.assertEqual(
                    getattr(module, status_attr), ModuleFileStatus.DONT_WANT
                )

        setattr(module, status_attr, ModuleFileStatus.DONT_NEED)
        self.assertEqual(getattr(module, status_attr), ModuleFileStatus.DONT_NEED)
        self.assertEqual(wants_file(), False)
        for status in set(ModuleFileStatus) - {
            ModuleFileStatus.WANT,
            ModuleFileStatus.DONT_WANT,
            ModuleFileStatus.DONT_NEED,
        }:
            with self.subTest(from_=ModuleFileStatus.DONT_NEED, to=status):
                self.assertRaises(ValueError, setattr, module, status_attr, status)
                self.assertEqual(
                    getattr(module, status_attr), ModuleFileStatus.DONT_NEED
                )

        setattr(module, status_attr, ModuleFileStatus.DONT_WANT)
        self.assertEqual(getattr(module, status_attr), ModuleFileStatus.DONT_WANT)

        setattr(module, status_attr, ModuleFileStatus.WANT)
        self.assertEqual(getattr(module, status_attr), ModuleFileStatus.WANT)

        setattr(module, status_attr, ModuleFileStatus.DONT_NEED)
        self.assertEqual(getattr(module, status_attr), ModuleFileStatus.DONT_NEED)

        setattr(module, status_attr, ModuleFileStatus.WANT)
        self.assertEqual(getattr(module, status_attr), ModuleFileStatus.WANT)

        self.assertRaises(AttributeError, delattr, module, status_attr)

    def test_loaded_file_status(self):
        self._test_file_status("loaded")

    def test_debug_file_status(self):
        self._test_file_status("debug")


class TestCreatedModules(TestCase):
    def test_empty(self):
        self.assertEqual(list(Program().modules()), [])

    def test_one(self):
        module = Program().extra_module("/foo/bar", create=True)
        self.assertEqual(list(module.prog.modules()), [module])

    def test_multiple(self):
        prog = Program()
        modules = [
            prog.extra_module("/foo/bar", create=True),
            prog.extra_module("/asdf/jkl", create=True),
            prog.extra_module("/123/456", create=True),
        ]
        self.assertCountEqual(list(prog.modules()), modules)

    def test_same_name(self):
        prog = Program()
        modules = [
            prog.extra_module("foo", id=0, create=True),
            prog.main_module("foo", create=True),
        ]
        actual = list(prog.modules())
        self.assertCountEqual(actual, modules)
        self.assertEqual(actual[0], prog.main_module())

        modules.append(prog.extra_module("foo", id=1, create=True))
        actual = list(prog.modules())
        self.assertCountEqual(actual, modules)
        self.assertEqual(actual[0], prog.main_module())

    def test_change_during_iteration(self):
        prog = Program()
        prog.extra_module("/foo/bar", create=True)
        with self.assertRaisesRegex(Exception, "modules changed during iteration"):
            for module in prog.modules():
                prog.extra_module("/asdf/jkl", create=True)
                prog.extra_module("/123/456", create=True)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import ast
import contextlib
import io
from pathlib import Path
import sys
import tempfile
import unittest.mock

from drgn.commands import (
    CommandArgumentError,
    CommandExitStatusError,
    CommandNamespace,
    CommandNotFoundError,
    DrgnCodeBuilder,
    ParsedCommand,
    _repr_black,
    _repr_raw_bytes,
    _sanitize_rst,
    argument,
    command,
    parse_shell_command,
    raw_command,
)
from tests import TestCase


class TestParseShellCommand(TestCase):
    def assert_parses_to(self, source, *, args, redirections=[], pipeline=None):
        self.assertEqual(
            parse_shell_command(source),
            ParsedCommand(args=args, redirections=redirections, pipeline=pipeline),
        )

    def test_one_arg(self):
        self.assert_parses_to("foo", args=["foo"])

    def test_multiple_args(self):
        self.assert_parses_to("foo bar   baz\t qux", args=["foo", "bar", "baz", "qux"])

    def test_escape_space(self):
        self.assert_parses_to(r"foo\ bar baz", args=["foo bar", "baz"])

    def test_escape_backslash(self):
        self.assert_parses_to(r"foo\\ bar", args=["foo\\", "bar"])

    def test_escape_double_quote(self):
        self.assert_parses_to(r"foo\"bar", args=['foo"bar'])

    def test_escape_single_quote(self):
        self.assert_parses_to(r"foo\'bar", args=["foo'bar"])

    def test_escape_other(self):
        self.assert_parses_to(r"foo\a", args=["fooa"])

    def test_double_quotes(self):
        self.assert_parses_to('"foo bar  baz\'qux"', args=["foo bar  baz'qux"])

    def test_double_quotes_empty(self):
        self.assert_parses_to('""', args=[""])

    def test_double_quotes_escape(self):
        # In double quotes, backslash only has special meaning before these
        # characters and newline.
        self.assert_parses_to(r' "\$\`\"\\" ', args=['$`"\\'])

    def test_double_quotes_escape_newline(self):
        self.assert_parses_to('"\\\n"', args=["\n"])

    def test_double_quotes_backslash_single_quote(self):
        self.assert_parses_to(r""" "\'" """, args=[r"\'"])

    def test_double_quotes_backslash_other(self):
        self.assert_parses_to(r'"\a"', args=[r"\a"])

    def test_single_quotes(self):
        self.assert_parses_to("' foo bar  \"baz'", args=[' foo bar  "baz'])

    def test_single_quotes_empty(self):
        self.assert_parses_to("''", args=[""])

    def test_single_quotes_backslash(self):
        # In single quotes, backslash does not have a special meaning.
        self.assert_parses_to(r"'\$'", args=[r"\$"])

    def test_single_quotes_backslash_single_quote(self):
        self.assert_parses_to(r"'\'", args=["\\"])

    def test_concatenate_escapes_and_quotes(self):
        self.assert_parses_to(r""" foo' bar'\ baz"q u x" """, args=["foo bar bazq u x"])

    def test_redirect_input(self):
        for source in (
            "foo < infile",
            "foo <infile",
            "foo< infile",
            "foo<infile",
            "<infile foo",
            "< infile foo",
        ):
            with self.subTest(source=source):
                self.assert_parses_to(
                    source, args=["foo"], redirections=[(0, "<", "infile")]
                )

    def test_redirect_input_fd(self):
        for source in (
            "foo 0< infile",
            "foo 0<infile",
            "0<infile foo",
            "0< infile foo",
        ):
            with self.subTest(source=source):
                self.assert_parses_to(
                    source, args=["foo"], redirections=[(0, "<", "infile")]
                )

    def test_redirect_output(self):
        for source in (
            "foo > outfile",
            "foo >outfile",
            "foo> outfile",
            "foo>outfile",
            ">outfile foo",
            "> outfile foo",
        ):
            with self.subTest(source=source):
                self.assert_parses_to(
                    source, args=["foo"], redirections=[(1, ">", "outfile")]
                )

    def test_redirect_output_fd(self):
        for source in (
            "foo 1> outfile",
            "foo 1>outfile",
            "1>outfile foo",
            "1> outfile foo",
        ):
            with self.subTest(source=source):
                self.assert_parses_to(
                    source, args=["foo"], redirections=[(1, ">", "outfile")]
                )

        for source in (
            "foo 2> outfile",
            "foo 2>outfile",
            "2>outfile foo",
            "2> outfile foo",
        ):
            with self.subTest(source=source):
                self.assert_parses_to(
                    source, args=["foo"], redirections=[(2, ">", "outfile")]
                )

    def test_append_output(self):
        for source in (
            "foo >> outfile",
            "foo >>outfile",
            "foo>> outfile",
            "foo>>outfile",
            ">>outfile foo",
            ">> outfile foo",
        ):
            with self.subTest(source=source):
                self.assert_parses_to(
                    source, args=["foo"], redirections=[(1, ">>", "outfile")]
                )

    def test_append_output_fd(self):
        for source in (
            "foo 1>> outfile",
            "foo 1>>outfile",
            "1>>outfile foo",
            "1>> outfile foo",
        ):
            with self.subTest(source=source):
                self.assert_parses_to(
                    source, args=["foo"], redirections=[(1, ">>", "outfile")]
                )

        for source in (
            "foo 2>> outfile",
            "foo 2>>outfile",
            "2>>outfile foo",
            "2>> outfile foo",
        ):
            with self.subTest(source=source):
                self.assert_parses_to(
                    source, args=["foo"], redirections=[(2, ">>", "outfile")]
                )

    def test_redirect_input_and_output(self):
        self.assert_parses_to(
            "foo < infile > outfile",
            args=["foo"],
            redirections=[(0, "<", "infile"), (1, ">", "outfile")],
        )

    def test_redirect_input_multiple(self):
        self.assert_parses_to(
            "< infile1 foo< infile2 <infile3",
            args=["foo"],
            redirections=[
                (0, "<", "infile1"),
                (0, "<", "infile2"),
                (0, "<", "infile3"),
            ],
        )

    def test_redirect_output_multiple(self):
        self.assert_parses_to(
            "> outfile1 foo>> outfile2 >outfile3",
            args=["foo"],
            redirections=[
                (1, ">", "outfile1"),
                (1, ">>", "outfile2"),
                (1, ">", "outfile3"),
            ],
        )

    def test_pipeline(self):
        self.assert_parses_to(
            "foo | grep 'bar baz' | sort",
            args=["foo"],
            pipeline="grep 'bar baz' | sort",
        )

    def test_pipeline_and_redirect(self):
        self.assert_parses_to(
            "foo < infile | grep 'bar baz' >outfile2",
            args=["foo"],
            redirections=[(0, "<", "infile")],
            pipeline="grep 'bar baz' >outfile2",
        )

    def test_only_redirect(self):
        self.assert_parses_to(
            "< infile >outfile",
            args=[],
            redirections=[(0, "<", "infile"), (1, ">", "outfile")],
        )

    def test_pipeline_empty(self):
        self.assertRaisesRegex(
            SyntaxError, "unexpected end of input", parse_shell_command, "foo | "
        )

    def test_pipe_after_redirect(self):
        for source in (
            "foo < |",
            "foo > |",
            "foo >> |",
        ):
            self.assertRaisesRegex(
                SyntaxError, r"unexpected '\|'", parse_shell_command, source
            )

    def test_nothing_before_pipe(self):
        self.assert_parses_to("| cat", args=[], pipeline="cat")

    def test_redirect_after_redirect(self):
        self.assertRaisesRegex(
            SyntaxError, "unexpected '<'", parse_shell_command, "foo < <"
        )
        self.assertRaisesRegex(
            SyntaxError, "unexpected '>'", parse_shell_command, "foo < >"
        )
        self.assertRaisesRegex(
            SyntaxError, "unexpected '>>'", parse_shell_command, "foo < >>"
        )

    def test_nothing_after_redirect(self):
        self.assertRaisesRegex(
            SyntaxError, "unexpected end of input", parse_shell_command, ">"
        )
        self.assertRaisesRegex(
            SyntaxError, "unexpected end of input", parse_shell_command, "foo >"
        )

    def test_comment(self):
        self.assert_parses_to("foo # bar < infile | baz", args=["foo"])

    def test_hash_in_word(self):
        self.assert_parses_to("foo#bar #baz", args=["foo#bar"])


_COMMANDS = CommandNamespace()


@command(
    description="write arguments to standard output",
    arguments=(argument("string", nargs="*"),),
    namespace=_COMMANDS,
)
def _cmd_echo(prog, name, args, **kwargs):
    print(*args.string)


@command(
    description="write arguments to standard error",
    arguments=(argument("string", nargs="*"),),
    namespace=_COMMANDS,
)
def _cmd_echoerr(prog, name, args, **kwargs):
    print(*args.string, file=sys.stderr)


@command(
    description="reverse standard input",
    namespace=_COMMANDS,
)
def _cmd_tac(prog, name, args, **kwargs):
    lines = sys.stdin.readlines()
    lines.reverse()
    sys.stdout.writelines(lines)


@raw_command(
    description="call ast.literal_eval()",
    usage="literal expr",
    long_description="",
    namespace=_COMMANDS,
)
def _cmd_literal(prog, name, args, **kwargs):
    return ast.literal_eval(args)


class TestRunCommand(TestCase):
    @staticmethod
    def run_command(source, **kwargs):
        return _COMMANDS.run(None, source, globals={}, **kwargs)

    def test_command_not_found(self):
        self.assertRaises(CommandNotFoundError, self.run_command, "asdf")

    def test_redirect_output(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir) / "file"
            self.run_command(f"echo foo bar > {path}")
            self.assertEqual(path.read_text(), "foo bar\n")

    def test_clobber_output(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir) / "file"
            path.write_text("x\n")
            self.run_command(f"echo foo bar > {path}")
            self.assertEqual(path.read_text(), "foo bar\n")

    def test_append_output(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir) / "file"
            path.write_text("foo\n")
            self.run_command(f"echo bar >> {path}")
            self.assertEqual(path.read_text(), "foo\nbar\n")

    def test_redirect_output_multiple(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path1 = Path(temp_dir) / "file1"
            path2 = Path(temp_dir) / "file2"
            self.run_command(f"echo foo bar > {path1} > {path2}")
            # Both files are created, but the output is written to the last
            # one.
            self.assertEqual(path1.read_text(), "")
            self.assertEqual(path2.read_text(), "foo bar\n")

    def test_redirect_error(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir) / "file"
            self.run_command(f"echoerr foo bar 2> {path}")
            self.assertEqual(path.read_text(), "foo bar\n")

    def test_clobber_error(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir) / "file"
            path.write_text("x\n")
            self.run_command(f"echoerr foo bar 2> {path}")
            self.assertEqual(path.read_text(), "foo bar\n")

    def test_append_error(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir) / "file"
            path.write_text("foo\n")
            self.run_command(f"echoerr bar 2>> {path}")
            self.assertEqual(path.read_text(), "foo\nbar\n")

    def test_redirect_input(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir) / "file"
            path.write_text("foo\nbar\n")
            f = io.StringIO()
            with contextlib.redirect_stdout(f):
                self.run_command(f"tac < {path}")
            self.assertEqual(f.getvalue(), "bar\nfoo\n")

    def test_redirect_input_multiple(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path1 = Path(temp_dir) / "file1"
            path2 = Path(temp_dir) / "file2"
            path1.write_text("foo\n")
            path2.write_text("bar\n")
            f = io.StringIO()
            with contextlib.redirect_stdout(f):
                self.run_command(f"tac < {path1} < {path2}")
            # The last redirection wins.
            self.assertEqual(f.getvalue(), "bar\n")

    def test_pipe(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir) / "file"
            self.run_command(f"echo hello | grep -o hell > {path}")
            self.assertEqual(path.read_text(), "hell\n")

    def test_redirect_input_and_pipe(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            in_path = Path(temp_dir) / "infile"
            out_path = Path(temp_dir) / "outfile"
            in_path.write_text("foo\nbar\n")
            self.run_command(f"tac < {in_path} | grep foo > {out_path}")
            self.assertEqual(out_path.read_text(), "foo\n")

    def test_redirect_output_and_pipe(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path1 = Path(temp_dir) / "file1"
            path2 = Path(temp_dir) / "file2"
            self.run_command(f"echo foo bar > {path1} | cat > {path2}")
            # Redirection wins over pipe.
            self.assertEqual(path1.read_text(), "foo bar\n")
            self.assertEqual(path2.read_text(), "")

    def test_redirect_input_failure(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir) / "file"
            self.assertRaises(OSError, self.run_command, f"tac < {path}")

    def test_redirect_output_failure(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path1 = Path(temp_dir) / "file"
            path2 = Path(temp_dir) / "dir/file"
            path1.write_text("foo\n")
            self.assertRaises(OSError, self.run_command, f"echo < {path1} > {path2}")

    def test_empty(self):
        self.assertRaisesRegex(
            SyntaxError, "expected command name", self.run_command, ""
        )
        self.assertRaisesRegex(
            SyntaxError, "expected command name", self.run_command, " \t"
        )
        self.assertRaisesRegex(
            SyntaxError, "expected command name", self.run_command, "# comment"
        )

    def test_not_command_name(self):
        self.assertRaisesRegex(
            SyntaxError, "expected command name", self.run_command, "< /dev/null"
        )
        self.assertRaisesRegex(
            SyntaxError, "expected command name", self.run_command, "| cat"
        )

    def test_redirect_invalid_input_fd(self):
        self.assertRaises(NotImplementedError, self.run_command, "echo 1< /dev/null")

    def test_redirect_invalid_output_fd(self):
        self.assertRaises(NotImplementedError, self.run_command, "echo 0> /dev/null")
        self.assertRaises(NotImplementedError, self.run_command, "echo 0>> /dev/null")

    def test_pipe_failure(self):
        self.assertRaises(CommandExitStatusError, self.run_command, "echo foo | false")

    def test_onerror(self):
        onerror = unittest.mock.Mock()
        self.run_command("asdf", onerror=onerror)
        onerror.assert_called_once()

    def test_onerror_raises(self):
        def onerror(e):
            1 / 0

        self.assertRaises(ZeroDivisionError, self.run_command, "asdf", onerror=onerror)

    def test_raw(self):
        self.assertEqual(self.run_command("literal (1, 2)"), (1, 2))

    def test_argument_error(self):
        self.assertRaises(
            CommandArgumentError, self.run_command, "tac --garbage 2> /dev/null"
        )


class TestSanitizeRst(TestCase):
    def test_empty(self):
        self.assertEqual(_sanitize_rst(""), "")

    def test_none(self):
        self.assertEqual(_sanitize_rst(None), None)

    def test_bold(self):
        self.assertEqual(_sanitize_rst("**true**"), "true")

    def test_italic(self):
        self.assertEqual(_sanitize_rst("*CPU*"), "CPU")

    def test_escaped_asterisk(self):
        self.assertEqual(_sanitize_rst(r"\*"), "*")

    def test_bold_escaped_asterisk(self):
        self.assertEqual(_sanitize_rst(r"**\***"), "*")

    def test_escaped_backslash(self):
        self.assertEqual(_sanitize_rst("\\\\"), "\\")

    def test_bold_and_escaped_dash(self):
        self.assertEqual(_sanitize_rst(r"**\-\-drgn**"), "--drgn")


class TestReprBlack(TestCase):
    def test_simple(self):
        self.assertEqual(_repr_black("foo"), '"foo"')

    def test_single_quote(self):
        self.assertEqual(_repr_black("ne'er"), '"ne\'er"')

    def test_double_quote(self):
        self.assertEqual(_repr_black('l"oL'), "'l\"oL'")


class TestReprRawBytes(TestCase):
    def test_simple(self):
        self.assertEqual(_repr_raw_bytes(b"foo"), 'rb"foo"')

    def test_single_quote(self):
        self.assertEqual(_repr_raw_bytes(b"ne'er"), 'rb"ne\'er"')

    def test_double_quote(self):
        self.assertEqual(_repr_raw_bytes(b'l"oL'), "rb'l\"oL'")

    def test_single_and_double_quote(self):
        self.assertEqual(_repr_raw_bytes(b"ne'er l\"oL"), 'b"ne\'er l\\"oL"')

    def test_ends_with_backslash(self):
        self.assertEqual(_repr_raw_bytes(b"foo\\"), r'b"foo\\"')

    def test_single_quote_and_ends_with_backslash(self):
        self.assertEqual(_repr_raw_bytes(b"ne'er\\"), 'b"ne\'er\\\\"')

    def test_double_quote_and_ends_with_backslash(self):
        self.assertEqual(_repr_raw_bytes(b'l"oL\\'), "b'l\"oL\\\\'")

    def test_single_and_double_quote_and_ends_with_backslash(self):
        self.assertEqual(_repr_raw_bytes(b"ne'er l\"oL\\"), 'b"ne\'er l\\"oL\\\\"')


class TestDrgnCodeBuilder(TestCase):
    def test_empty(self):
        self.assertEqual(DrgnCodeBuilder(None).get(), "")

    def test_no_imports(self):
        code = DrgnCodeBuilder(None)
        code.append("pass\n")
        self.assertEqual(code.get(), "pass\n")

    def test_sort_imports(self):
        code = DrgnCodeBuilder(None)
        code.add_import("sys")
        code.add_import("os")
        self.assertEqual(
            code.get(),
            """\
import os
import sys
""",
        )

    def test_deduplicate_imports(self):
        code = DrgnCodeBuilder(None)
        code.add_import("os")
        code.add_import("sys")
        code.add_import("os")
        self.assertEqual(
            code.get(),
            """\
import os
import sys
""",
        )

    def test_sort_from_imports(self):
        code = DrgnCodeBuilder(None)
        code.add_import("sys")
        code.add_import("os")
        code.add_from_import("sys", "stdout")
        self.assertEqual(
            code.get(),
            """\
import os
import sys
from sys import stdout
""",
        )

    def test_first_party_imports(self):
        code = DrgnCodeBuilder(None)
        code.add_import("os")
        code.add_import("sys")
        code.add_import("drgn")
        self.assertEqual(
            code.get(),
            """\
import os
import sys

import drgn
""",
        )

    def test_merge_from_imports(self):
        code = DrgnCodeBuilder(None)
        code.add_from_import("drgn", "Object", "Program")
        code.add_from_import("drgn", "Object")
        code.add_from_import("drgn", "Type")
        self.assertEqual(
            code.get(),
            """\
from drgn import Object, Program, Type
""",
        )

    def test_long_from_imports(self):
        code = DrgnCodeBuilder(None)
        code.add_from_import(
            "os",
            "abort",
            "access",
            "chdir",
            "chmod",
            "chown",
            "close",
            "closerange",
            "confstr",
            "copy_file_range",
        )
        self.assertEqual(
            code.get(),
            """\
from os import (
    abort,
    access,
    chdir,
    chmod,
    chown,
    close,
    closerange,
    confstr,
    copy_file_range,
)
""",
        )

    def test_imports_and_code(self):
        code = DrgnCodeBuilder(None)
        code.append("pass\n")
        code.add_import("os")
        self.assertEqual(
            code.get(),
            """\
import os


pass
""",
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
from pathlib import Path
import subprocess
import tempfile
import unittest

from util import out_of_date


def get_resource(name: str) -> Path:
    dir = Path(__file__).parent
    decompressed_path = dir / name
    compressed_path = dir / (name + ".zst")
    if out_of_date(decompressed_path, compressed_path):
        tmp_file = tempfile.NamedTemporaryFile(dir=dir, prefix=name, delete=False)
        try:
            try:
                subprocess.check_call(
                    [
                        "zstd",
                        "--quiet",
                        "--force",
                        "--decompress",
                        "--stdout",
                        str(compressed_path),
                    ],
                    stdout=tmp_file,
                )
            except FileNotFoundError:
                raise unittest.SkipTest("zstd not found")
        except BaseException:
            os.unlink(tmp_file.name)
            raise
        else:
            os.rename(tmp_file.name, decompressed_path)
    return decompressed_path



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import json
import os
from pathlib import Path
import typing
from typing import Any, Dict, Mapping, Optional, Union
import urllib.error
import urllib.parse
import urllib.request

if typing.TYPE_CHECKING:
    import aiohttp


_CACHE = Optional[Union[str, bytes, Path]]


# Hacky base class because we want the GitHub API from async and non-async
# code.
#
# This provides a slapdash interface for caching a response in a file so that
# we can do conditional requests
# (https://docs.github.com/en/rest/overview/resources-in-the-rest-api#conditional-requests).
# A more complete implementation would be something like a SQLite database
# indexed by endpoint, but this is simpler and good enough for now.
class _GitHubApiBase:
    _HOST = "https://api.github.com"

    def __init__(self, token: Optional[str]) -> None:
        self._headers = {
            "Accept": "application/vnd.github.v3+json",
            "User-Agent": "osandov/drgn vmtest",
        }
        if token is not None:
            self._headers["Authorization"] = "token " + token

    def _request(
        self,
        method: str,
        url: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        data: Any = None,
    ) -> Any:
        raise NotImplementedError()

    def _cached_get_json(self, endpoint: str, cache: _CACHE) -> Any:
        raise NotImplementedError()

    def _read_cache(self, cache: _CACHE) -> Optional[Mapping[str, Any]]:
        if not cache:
            return None
        try:
            with open(cache, "r") as f:
                return json.load(f)
        except FileNotFoundError:
            return None

    def _cached_get_headers(
        self, cached: Optional[Mapping[str, Any]]
    ) -> Dict[str, str]:
        if cached is not None:
            if "etag" in cached:
                return {**self._headers, "If-None-Match": cached["etag"]}
            elif "last_modified" in cached:
                return {**self._headers, "If-Modified-Since": cached["last_modified"]}
        return self._headers

    def _write_cache(
        self, cache: _CACHE, body: Any, headers: Mapping[str, str]
    ) -> None:
        if cache is not None and ("ETag" in headers or "Last-Modified" in headers):
            to_cache = {"body": body}
            if "ETag" in headers:
                to_cache["etag"] = headers["ETag"]
            if "Last-Modified" in headers:
                to_cache["last_modified"] = headers["Last-Modified"]
            with open(cache, "w") as f:
                json.dump(to_cache, f)

    def get_release_by_tag(
        self, owner: str, repo: str, tag: str, *, cache: _CACHE = None
    ) -> Any:
        return self._cached_get_json(f"repos/{owner}/{repo}/releases/tags/{tag}", cache)

    def download(self, url: str) -> Any:
        return self._request(
            "GET", url, headers={**self._headers, "Accept": "application/octet-stream"}
        )

    def upload(
        self,
        url: str,
        data: Any,
        content_type: str,
        content_length: Optional[int] = None,
    ) -> Any:
        headers = {**self._headers, "Content-Type": content_type}
        if content_length is not None:
            headers["Content-Length"] = str(content_length)
        return self._request(
            "POST",
            url,
            headers=headers,
            data=data,
        )

    def delete(self, url: str) -> Any:
        return self._request("DELETE", url, headers=self._headers)


class GitHubApi(_GitHubApiBase):
    def _request(
        self,
        method: str,
        url: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        data: Any = None,
    ) -> Any:
        if params:
            url += "?" + urllib.parse.urlencode(params)
        req = urllib.request.Request(
            url,
            data=data,
            headers={} if headers is None else headers,
            method=method,
        )
        # Work around python/cpython#77842.
        authorization = req.get_header("Authorization")
        if authorization is not None:
            req.remove_header("Authorization")
            req.add_unredirected_header("Authorization", authorization)
        return urllib.request.urlopen(req)

    def _cached_get_json(self, endpoint: str, cache: _CACHE) -> Any:
        cached = self._read_cache(cache)
        # If the request was cached and the VMTEST_TRUST_CACHE environment
        # variable is set, assume the cache is still valid.
        if cached is not None and "VMTEST_TRUST_CACHE" in os.environ:
            return cached["body"]
        req = urllib.request.Request(
            self._HOST + "/" + endpoint,
            headers=self._cached_get_headers(cached),
        )
        # Work around python/cpython#77842.
        authorization = req.get_header("Authorization")
        if authorization is not None:
            req.remove_header("Authorization")
            req.add_unredirected_header("Authorization", authorization)
        try:
            with urllib.request.urlopen(req) as resp:
                body = json.load(resp)
                self._write_cache(cache, body, resp.headers)
                return body
        except urllib.error.HTTPError as e:
            if e.code == 304 and cached is not None:
                return cached["body"]
            else:
                raise


class AioGitHubApi(_GitHubApiBase):
    def __init__(self, session: "aiohttp.ClientSession", token: Optional[str]) -> None:
        super().__init__(token)
        self._session = session

    def _request(
        self,
        method: str,
        url: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        data: Any = None,
    ) -> Any:
        return self._session.request(
            method,
            url,
            params=params,
            headers=headers,
            data=data,
            raise_for_status=True,
        )

    async def _cached_get_json(self, endpoint: str, cache: _CACHE) -> Any:
        cached = self._read_cache(cache)
        if cached is not None and "VMTEST_TRUST_CACHE" in os.environ:
            return cached["body"]
        async with self._session.get(
            self._HOST + "/" + endpoint,
            headers=self._cached_get_headers(cached),
            raise_for_status=True,
        ) as resp:
            if resp.status == 304:
                if cached is None:
                    raise Exception("got HTTP 304 but response was not cached")
                return cached["body"]
            else:
                body = await resp.json()
                self._write_cache(cache, body, resp.headers)
                return body



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn import (
    Architecture,
    Language,
    NoDefaultProgramError,
    Object,
    Platform,
    PrimitiveType,
    Program,
    Qualifiers,
    TypeEnumerator,
    TypeKind,
    TypeMember,
    TypeParameter,
    TypeTemplateParameter,
    alignof,
    offsetof,
    sizeof,
)
from tests import (
    DEFAULT_LANGUAGE,
    MockProgramTestCase,
    TestCase,
    assertReprPrettyEqualsStr,
    mock_program,
    with_default_prog,
)


class TestType(MockProgramTestCase):
    def test_void(self):
        t = self.prog.void_type()
        self.assertEqual(t.kind, TypeKind.VOID)
        self.assertEqual(t.primitive, PrimitiveType.C_VOID)
        self.assertEqual(t.language, DEFAULT_LANGUAGE)
        self.assertIdentical(t, self.prog.void_type())
        self.assertFalse(t.is_complete())
        self.assertEqual(repr(t), "prog.void_type()")

    def test_int(self):
        t = self.prog.int_type("int", 4, True)
        self.assertEqual(t.kind, TypeKind.INT)
        self.assertEqual(t.primitive, PrimitiveType.C_INT)
        self.assertEqual(t.language, DEFAULT_LANGUAGE)
        self.assertEqual(t.name, "int")
        self.assertEqual(t.size, 4)
        self.assertTrue(t.is_signed)
        self.assertEqual(t.byteorder, "little")
        self.assertTrue(t.is_complete())

        self.assertEqual(repr(t), "prog.int_type(name='int', size=4, is_signed=True)")
        self.assertEqual(sizeof(t), 4)

        self.assertRaises(TypeError, self.prog.int_type, None, 4, True)

        self.assertIsNone(self.prog.int_type("my_int", 4, True).primitive)
        self.assertIsNone(self.prog.int_type("int", 4, False).primitive)

    def test_int_byteorder(self):
        self.assertIdentical(
            self.prog.int_type("int", 4, True),
            self.prog.int_type("int", 4, True, byteorder="little"),
        )
        t = self.prog.int_type("int", 4, True, "big")
        self.assertEqual(t.byteorder, "big")
        self.assertEqual(
            repr(t),
            "prog.int_type(name='int', size=4, is_signed=True, byteorder='big')",
        )
        self.assertRaises(ValueError, self.prog.int_type, "int", 4, True, "middle")

    def test_bool(self):
        t = self.prog.bool_type("_Bool", 1)
        self.assertEqual(t.kind, TypeKind.BOOL)
        self.assertEqual(t.primitive, PrimitiveType.C_BOOL)
        self.assertEqual(t.language, DEFAULT_LANGUAGE)
        self.assertEqual(t.name, "_Bool")
        self.assertEqual(t.size, 1)
        self.assertEqual(t.byteorder, "little")
        self.assertTrue(t.is_complete())

        self.assertEqual(repr(t), "prog.bool_type(name='_Bool', size=1)")
        self.assertEqual(sizeof(t), 1)

        self.assertRaises(TypeError, self.prog.bool_type, None, 1)

    def test_bool_byteorder(self):
        self.assertIdentical(
            self.prog.bool_type("_Bool", 1),
            self.prog.bool_type("_Bool", 1, byteorder="little"),
        )
        t = self.prog.bool_type("_Bool", 1, "big")
        self.assertEqual(t.byteorder, "big")
        self.assertEqual(
            repr(t), "prog.bool_type(name='_Bool', size=1, byteorder='big')"
        )
        self.assertRaises(ValueError, self.prog.bool_type, "_Bool", 1, "middle")

    def test_float(self):
        t = self.prog.float_type("float", 4)
        self.assertEqual(t.primitive, PrimitiveType.C_FLOAT)
        self.assertEqual(t.kind, TypeKind.FLOAT)
        self.assertEqual(t.name, "float")
        self.assertEqual(t.size, 4)
        self.assertEqual(t.byteorder, "little")
        self.assertTrue(t.is_complete())

        self.assertEqual(repr(t), "prog.float_type(name='float', size=4)")
        self.assertEqual(sizeof(t), 4)

        self.assertRaises(TypeError, self.prog.float_type, None, 4)

    def test_float_byteorder(self):
        self.assertIdentical(
            self.prog.bool_type("float", 1),
            self.prog.bool_type("float", 1, byteorder="little"),
        )
        t = self.prog.float_type("float", 4, "big")
        self.assertEqual(t.byteorder, "big")
        self.assertEqual(
            repr(t), "prog.float_type(name='float', size=4, byteorder='big')"
        )
        self.assertRaises(ValueError, self.prog.float_type, "float", 4, "middle")

    def test_struct(self):
        t = self.prog.struct_type(
            "point",
            8,
            (
                TypeMember(self.prog.int_type("int", 4, True), "x", 0),
                TypeMember(self.prog.int_type("int", 4, True), "y", 32),
            ),
        )
        self.assertEqual(t.kind, TypeKind.STRUCT)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.language, DEFAULT_LANGUAGE)
        self.assertEqual(t.tag, "point")
        self.assertEqual(t.size, 8)
        self.assertIdentical(
            t.members,
            (
                TypeMember(self.prog.int_type("int", 4, True), "x", 0),
                TypeMember(self.prog.int_type("int", 4, True), "y", 32),
            ),
        )
        self.assertTrue(t.is_complete())

        self.assertEqual(
            repr(t),
            "prog.struct_type(tag='point', size=8, members=(TypeMember(prog.type('int'), name='x', bit_offset=0), TypeMember(prog.type('int'), name='y', bit_offset=32)))",
        )
        self.assertEqual(sizeof(t), 8)

        t = self.prog.struct_type(
            None,
            8,
            (
                TypeMember(self.prog.int_type("int", 4, True), "x", 0),
                TypeMember(self.prog.int_type("int", 4, True), "y", 32),
            ),
        )
        self.assertEqual(t.kind, TypeKind.STRUCT)
        self.assertIsNone(t.primitive)
        self.assertIsNone(t.tag)
        self.assertEqual(t.size, 8)
        self.assertIdentical(
            t.members,
            (
                TypeMember(self.prog.int_type("int", 4, True), "x", 0),
                TypeMember(self.prog.int_type("int", 4, True), "y", 32),
            ),
        )
        self.assertTrue(t.is_complete())

        t = self.prog.struct_type("color", 0, ())
        self.assertEqual(t.kind, TypeKind.STRUCT)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.tag, "color")
        self.assertEqual(t.size, 0)
        self.assertEqual(t.members, ())
        self.assertTrue(t.is_complete())
        self.assertEqual(repr(t), "prog.struct_type(tag='color', size=0, members=())")

        t = self.prog.struct_type("color")
        self.assertEqual(t.kind, TypeKind.STRUCT)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.tag, "color")
        self.assertIsNone(t.size)
        self.assertIsNone(t.members)
        self.assertFalse(t.is_complete())
        self.assertEqual(
            repr(t), "prog.struct_type(tag='color', size=None, members=None)"
        )

        t = self.prog.struct_type(None, None, None)
        self.assertEqual(t.kind, TypeKind.STRUCT)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.tag, None)
        self.assertIsNone(t.size)
        self.assertIsNone(t.members)
        self.assertFalse(t.is_complete())
        self.assertEqual(repr(t), "prog.struct_type(tag=None, size=None, members=None)")

        self.assertRaises(TypeError, self.prog.struct_type, 4)
        self.assertRaisesRegex(
            ValueError, "must not have size", self.prog.struct_type, "point", 8, None
        )
        self.assertRaisesRegex(
            ValueError, "must have size", self.prog.struct_type, "point", None, ()
        )
        self.assertRaisesRegex(
            TypeError, "must be sequence or None", self.prog.struct_type, "point", 8, 4
        )
        self.assertRaisesRegex(
            TypeError, "must be TypeMember", self.prog.struct_type, "point", 8, (4,)
        )

        # Bit size.
        t = self.prog.struct_type(
            "point",
            8,
            (
                TypeMember(
                    Object(
                        self.prog, self.prog.int_type("int", 4, True), bit_field_size=4
                    ),
                    "x",
                    0,
                ),
                TypeMember(
                    Object(
                        self.prog, self.prog.int_type("int", 4, True), bit_field_size=4
                    ),
                    "y",
                    32,
                ),
            ),
        )
        self.assertIdentical(
            t.members,
            (
                TypeMember(
                    Object(
                        self.prog, self.prog.int_type("int", 4, True), bit_field_size=4
                    ),
                    "x",
                    0,
                ),
                TypeMember(
                    Object(
                        self.prog, self.prog.int_type("int", 4, True), bit_field_size=4
                    ),
                    "y",
                    32,
                ),
            ),
        )

    def test_union(self):
        t = self.prog.union_type(
            "option",
            4,
            (
                TypeMember(self.prog.int_type("int", 4, True), "x"),
                TypeMember(self.prog.int_type("unsigned int", 4, False), "y"),
            ),
        )
        self.assertEqual(t.kind, TypeKind.UNION)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.language, DEFAULT_LANGUAGE)
        self.assertEqual(t.tag, "option")
        self.assertEqual(t.size, 4)
        self.assertIdentical(
            t.members,
            (
                TypeMember(self.prog.int_type("int", 4, True), "x", 0),
                TypeMember(self.prog.int_type("unsigned int", 4, False), "y", 0),
            ),
        )
        self.assertTrue(t.is_complete())

        self.assertEqual(
            repr(t),
            "prog.union_type(tag='option', size=4, members=(TypeMember(prog.type('int'), name='x', bit_offset=0), TypeMember(prog.type('unsigned int'), name='y', bit_offset=0)))",
        )
        self.assertEqual(sizeof(t), 4)

        t = self.prog.union_type(
            None,
            4,
            (
                TypeMember(self.prog.int_type("int", 4, True), "x"),
                TypeMember(self.prog.int_type("unsigned int", 4, False), "y"),
            ),
        )
        self.assertEqual(t.kind, TypeKind.UNION)
        self.assertIsNone(t.primitive)
        self.assertIsNone(t.tag)
        self.assertEqual(t.size, 4)
        self.assertIdentical(
            t.members,
            (
                TypeMember(self.prog.int_type("int", 4, True), "x", 0),
                TypeMember(self.prog.int_type("unsigned int", 4, False), "y", 0),
            ),
        )
        self.assertTrue(t.is_complete())

        t = self.prog.union_type("color", 0, ())
        self.assertEqual(t.kind, TypeKind.UNION)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.tag, "color")
        self.assertEqual(t.size, 0)
        self.assertEqual(t.members, ())
        self.assertTrue(t.is_complete())
        self.assertEqual(repr(t), "prog.union_type(tag='color', size=0, members=())")

        t = self.prog.union_type("color")
        self.assertEqual(t.kind, TypeKind.UNION)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.tag, "color")
        self.assertIsNone(t.size)
        self.assertIsNone(t.members)
        self.assertFalse(t.is_complete())
        self.assertEqual(
            repr(t), "prog.union_type(tag='color', size=None, members=None)"
        )

        t = self.prog.union_type(None, None, None)
        self.assertEqual(t.kind, TypeKind.UNION)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.tag, None)
        self.assertIsNone(t.size)
        self.assertIsNone(t.members)
        self.assertFalse(t.is_complete())
        self.assertEqual(repr(t), "prog.union_type(tag=None, size=None, members=None)")

        self.assertRaises(TypeError, self.prog.union_type, 4)
        self.assertRaisesRegex(
            ValueError, "must not have size", self.prog.union_type, "option", 8, None
        )
        self.assertRaisesRegex(
            ValueError, "must have size", self.prog.union_type, "option", None, ()
        )
        self.assertRaisesRegex(
            TypeError, "must be sequence or None", self.prog.union_type, "option", 8, 4
        )
        self.assertRaisesRegex(
            TypeError, "must be TypeMember", self.prog.union_type, "option", 8, (4,)
        )

        # Bit size.
        t = self.prog.union_type(
            "option",
            4,
            (
                TypeMember(
                    Object(
                        self.prog, self.prog.int_type("int", 4, True), bit_field_size=4
                    ),
                    "x",
                    0,
                ),
                TypeMember(
                    Object(
                        self.prog,
                        self.prog.int_type("unsigned int", 4, False),
                        bit_field_size=4,
                    ),
                    "y",
                    0,
                ),
            ),
        )
        self.assertIdentical(
            t.members,
            (
                TypeMember(
                    Object(
                        self.prog, self.prog.int_type("int", 4, True), bit_field_size=4
                    ),
                    "x",
                    0,
                ),
                TypeMember(
                    Object(
                        self.prog,
                        self.prog.int_type("unsigned int", 4, False),
                        bit_field_size=4,
                    ),
                    "y",
                    0,
                ),
            ),
        )

    def test_class(self):
        t = self.prog.class_type(
            "coord",
            12,
            (
                TypeMember(self.prog.int_type("int", 4, True), "x", 0),
                TypeMember(self.prog.int_type("int", 4, True), "y", 32),
                TypeMember(self.prog.int_type("int", 4, True), "z", 64),
            ),
        )
        self.assertEqual(t.kind, TypeKind.CLASS)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.language, DEFAULT_LANGUAGE)
        self.assertEqual(t.tag, "coord")
        self.assertEqual(t.size, 12)
        self.assertIdentical(
            t.members,
            (
                TypeMember(self.prog.int_type("int", 4, True), "x", 0),
                TypeMember(self.prog.int_type("int", 4, True), "y", 32),
                TypeMember(self.prog.int_type("int", 4, True), "z", 64),
            ),
        )
        self.assertTrue(t.is_complete())

        self.assertEqual(
            repr(t),
            "prog.class_type(tag='coord', size=12, members=(TypeMember(prog.type('int'), name='x', bit_offset=0), TypeMember(prog.type('int'), name='y', bit_offset=32), TypeMember(prog.type('int'), name='z', bit_offset=64)))",
        )
        self.assertEqual(sizeof(t), 12)

        t = self.prog.class_type(
            None,
            12,
            (
                TypeMember(self.prog.int_type("int", 4, True), "x", 0),
                TypeMember(self.prog.int_type("int", 4, True), "y", 32),
                TypeMember(self.prog.int_type("int", 4, True), "z", 64),
            ),
        )
        self.assertEqual(t.kind, TypeKind.CLASS)
        self.assertIsNone(t.primitive)
        self.assertIsNone(t.tag)
        self.assertEqual(t.size, 12)
        self.assertIdentical(
            t.members,
            (
                TypeMember(self.prog.int_type("int", 4, True), "x", 0),
                TypeMember(self.prog.int_type("int", 4, True), "y", 32),
                TypeMember(self.prog.int_type("int", 4, True), "z", 64),
            ),
        )
        self.assertTrue(t.is_complete())

        t = self.prog.class_type("color", 0, ())
        self.assertEqual(t.kind, TypeKind.CLASS)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.tag, "color")
        self.assertEqual(t.size, 0)
        self.assertEqual(t.members, ())
        self.assertTrue(t.is_complete())
        self.assertEqual(repr(t), "prog.class_type(tag='color', size=0, members=())")

        t = self.prog.class_type("color")
        self.assertEqual(t.kind, TypeKind.CLASS)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.tag, "color")
        self.assertIsNone(t.size)
        self.assertIsNone(t.members)
        self.assertFalse(t.is_complete())
        self.assertEqual(
            repr(t), "prog.class_type(tag='color', size=None, members=None)"
        )

        t = self.prog.class_type(None, None, None)
        self.assertEqual(t.kind, TypeKind.CLASS)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.tag, None)
        self.assertIsNone(t.size)
        self.assertIsNone(t.members)
        self.assertFalse(t.is_complete())
        self.assertEqual(repr(t), "prog.class_type(tag=None, size=None, members=None)")

        self.assertRaises(TypeError, self.prog.class_type, 4)
        self.assertRaisesRegex(
            ValueError, "must not have size", self.prog.class_type, "coord", 12, None
        )
        self.assertRaisesRegex(
            ValueError, "must have size", self.prog.class_type, "coord", None, ()
        )
        self.assertRaisesRegex(
            TypeError, "must be sequence or None", self.prog.class_type, "coord", 12, 4
        )
        self.assertRaisesRegex(
            TypeError, "must be TypeMember", self.prog.class_type, "coord", 12, (4,)
        )

        # Bit size.
        t = self.prog.class_type(
            "coord",
            12,
            (
                TypeMember(
                    Object(
                        self.prog, self.prog.int_type("int", 4, True), bit_field_size=4
                    ),
                    "x",
                    0,
                ),
                TypeMember(
                    Object(
                        self.prog, self.prog.int_type("int", 4, True), bit_field_size=4
                    ),
                    "y",
                    32,
                ),
                TypeMember(
                    Object(
                        self.prog, self.prog.int_type("int", 4, True), bit_field_size=4
                    ),
                    "z",
                    64,
                ),
            ),
        )
        self.assertIdentical(
            t.members,
            (
                TypeMember(
                    Object(
                        self.prog, self.prog.int_type("int", 4, True), bit_field_size=4
                    ),
                    "x",
                    0,
                ),
                TypeMember(
                    Object(
                        self.prog, self.prog.int_type("int", 4, True), bit_field_size=4
                    ),
                    "y",
                    32,
                ),
                TypeMember(
                    Object(
                        self.prog, self.prog.int_type("int", 4, True), bit_field_size=4
                    ),
                    "z",
                    64,
                ),
            ),
        )

    def test_member(self):
        t = self.prog.struct_type(
            None,
            8,
            (
                TypeMember(self.prog.int_type("int", 4, True), "x", 0),
                TypeMember(
                    self.prog.struct_type(
                        None,
                        4,
                        (TypeMember(self.prog.int_type("int", 4, True), "y", 0),),
                    ),
                    None,
                    32,
                ),
            ),
        )
        self.assertIdentical(
            t.member("x"), TypeMember(self.prog.int_type("int", 4, True), "x", 0)
        )
        self.assertIdentical(
            t.member("y"), TypeMember(self.prog.int_type("int", 4, True), "y", 32)
        )
        self.assertRaises(LookupError, t.member, "z")

        self.assertIdentical(
            t.members[1].type.member("y"),
            TypeMember(self.prog.int_type("int", 4, True), "y", 0),
        )

        self.assertRaises(TypeError, self.prog.int_type("int", 4, True).member, "foo")

    def test_offsetof(self):
        self.assertEqual(offsetof(self.line_segment_type, "b"), 8)
        self.assertEqual(offsetof(self.line_segment_type, "a.y"), 4)
        self.assertRaisesRegex(
            LookupError,
            "'struct line_segment' has no member 'c'",
            offsetof,
            self.line_segment_type,
            "c.x",
        )

        small_point_type = self.prog.struct_type(
            "small_point",
            1,
            (
                TypeMember(
                    Object(
                        self.prog, self.prog.int_type("int", 4, True), bit_field_size=4
                    ),
                    "x",
                    0,
                ),
                TypeMember(
                    Object(
                        self.prog, self.prog.int_type("int", 4, True), bit_field_size=4
                    ),
                    "y",
                    4,
                ),
            ),
        )
        self.assertEqual(offsetof(small_point_type, "x"), 0)
        self.assertRaisesRegex(
            ValueError,
            "member is not byte-aligned",
            offsetof,
            small_point_type,
            "y",
        )
        self.assertRaisesRegex(
            SyntaxError,
            "expected identifier",
            offsetof,
            self.prog.array_type(self.prog.int_type("int", 4, True), 10),
            "[1]",
        )

    def test_enum(self):
        t = self.prog.enum_type(
            "color",
            self.prog.int_type("unsigned int", 4, False),
            (
                TypeEnumerator("RED", 0),
                TypeEnumerator("GREEN", 1),
                TypeEnumerator("BLUE", 2),
            ),
        )
        self.assertEqual(t.kind, TypeKind.ENUM)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.language, DEFAULT_LANGUAGE)
        self.assertEqual(t.tag, "color")
        self.assertIdentical(t.type, self.prog.int_type("unsigned int", 4, False))
        self.assertEqual(
            t.enumerators,
            (
                TypeEnumerator("RED", 0),
                TypeEnumerator("GREEN", 1),
                TypeEnumerator("BLUE", 2),
            ),
        )
        self.assertTrue(t.is_complete())

        self.assertEqual(
            repr(t),
            "prog.enum_type(tag='color', type=prog.int_type(name='unsigned int', size=4, is_signed=False), enumerators=(TypeEnumerator('RED', 0), TypeEnumerator('GREEN', 1), TypeEnumerator('BLUE', 2)))",
        )
        self.assertEqual(sizeof(t), 4)

        t = self.prog.enum_type("color", None, None)
        self.assertEqual(t.kind, TypeKind.ENUM)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.tag, "color")
        self.assertIsNone(t.type)
        self.assertIsNone(t.enumerators)
        self.assertFalse(t.is_complete())

        self.assertEqual(
            repr(t), "prog.enum_type(tag='color', type=None, enumerators=None)"
        )

        # A type with no enumerators isn't valid in C, but we allow it.
        t = self.prog.enum_type(
            "color", self.prog.int_type("unsigned int", 4, False), ()
        )
        self.assertEqual(t.kind, TypeKind.ENUM)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.tag, "color")
        self.assertIdentical(t.type, self.prog.int_type("unsigned int", 4, False))
        self.assertEqual(t.enumerators, ())
        self.assertTrue(t.is_complete())

        self.assertEqual(
            repr(t),
            "prog.enum_type(tag='color', type=prog.int_type(name='unsigned int', size=4, is_signed=False), enumerators=())",
        )

        self.assertRaisesRegex(
            TypeError, "must be Type", self.prog.enum_type, "color", 4, ()
        )
        self.assertRaisesRegex(
            ValueError,
            "must be integer type",
            self.prog.enum_type,
            "color",
            self.prog.void_type(),
            (),
        )
        self.assertRaisesRegex(
            ValueError,
            "must be unqualified",
            self.prog.enum_type,
            "color",
            self.prog.int_type("unsigned int", 4, True, qualifiers=Qualifiers.CONST),
            (),
        )
        self.assertRaisesRegex(
            ValueError,
            "must not have compatible type",
            self.prog.enum_type,
            "color",
            self.prog.int_type("unsigned int", 4, False),
            None,
        )
        self.assertRaisesRegex(
            ValueError,
            "must have compatible type",
            self.prog.enum_type,
            "color",
            None,
            (),
        )
        self.assertRaisesRegex(
            TypeError,
            "must be sequence or None",
            self.prog.enum_type,
            "color",
            self.prog.int_type("unsigned int", 4, False),
            4,
        )
        self.assertRaisesRegex(
            TypeError,
            "must be TypeEnumerator",
            self.prog.enum_type,
            "color",
            self.prog.int_type("unsigned int", 4, False),
            (4,),
        )

    def test_typedef(self):
        t = self.prog.typedef_type("INT", self.prog.int_type("int", 4, True))
        self.assertEqual(t.kind, TypeKind.TYPEDEF)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.language, DEFAULT_LANGUAGE)
        self.assertEqual(t.name, "INT")
        self.assertIdentical(t.type, self.prog.int_type("int", 4, True))
        self.assertTrue(t.is_complete())

        self.assertEqual(
            repr(t),
            "prog.typedef_type(name='INT', type=prog.int_type(name='int', size=4, is_signed=True))",
        )
        self.assertEqual(sizeof(t), 4)

        t = self.prog.typedef_type("VOID", self.prog.void_type())
        self.assertFalse(t.is_complete())

        self.assertRaises(
            TypeError, self.prog.typedef_type, None, self.prog.int_type("int", 4, True)
        )
        self.assertRaises(TypeError, self.prog.typedef_type, "INT", 4)

        self.assertEqual(
            self.prog.typedef_type(
                "size_t", self.prog.int_type("unsigned long", 8, False)
            ).primitive,
            PrimitiveType.C_SIZE_T,
        )
        self.assertEqual(
            self.prog.typedef_type(
                "ptrdiff_t", self.prog.int_type("long", 8, True)
            ).primitive,
            PrimitiveType.C_PTRDIFF_T,
        )

    def test_pointer(self):
        t = self.prog.pointer_type(self.prog.int_type("int", 4, True), 8)
        self.assertEqual(t.kind, TypeKind.POINTER)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.language, DEFAULT_LANGUAGE)
        self.assertEqual(t.size, 8)
        self.assertIdentical(t.type, self.prog.int_type("int", 4, True))
        self.assertEqual(t.byteorder, "little")
        self.assertTrue(t.is_complete())

        self.assertEqual(
            repr(t),
            "prog.pointer_type(type=prog.int_type(name='int', size=4, is_signed=True))",
        )
        self.assertEqual(
            repr(self.prog.pointer_type(self.prog.int_type("int", 4, True), 4)),
            "prog.pointer_type(type=prog.int_type(name='int', size=4, is_signed=True), size=4)",
        )

        self.assertEqual(sizeof(t), 8)

        self.assertRaises(TypeError, self.prog.pointer_type, 4)

    def test_pointer_byteorder(self):
        self.assertIdentical(
            self.prog.pointer_type(self.prog.int_type("int", 4, True)),
            self.prog.pointer_type(
                self.prog.int_type("int", 4, True), byteorder="little"
            ),
        )
        t = self.prog.pointer_type(self.prog.int_type("int", 4, True), 8, "big")
        self.assertEqual(t.byteorder, "big")
        self.assertEqual(t.type.byteorder, "little")
        self.assertEqual(
            repr(t),
            "prog.pointer_type(type=prog.int_type(name='int', size=4, is_signed=True), byteorder='big')",
        )
        self.assertRaises(
            ValueError,
            self.prog.pointer_type,
            self.prog.int_type("int", 4, True),
            8,
            "middle",
        )

    def test_array(self):
        t = self.prog.array_type(self.prog.int_type("int", 4, True), 10)
        self.assertEqual(t.kind, TypeKind.ARRAY)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.language, DEFAULT_LANGUAGE)
        self.assertEqual(t.length, 10)
        self.assertIdentical(t.type, self.prog.int_type("int", 4, True))
        self.assertTrue(t.is_complete())

        self.assertEqual(
            repr(t),
            "prog.array_type(type=prog.int_type(name='int', size=4, is_signed=True), length=10)",
        )
        self.assertEqual(sizeof(t), 40)

        t = self.prog.array_type(self.prog.int_type("int", 4, True), 0)
        self.assertEqual(t.kind, TypeKind.ARRAY)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.length, 0)
        self.assertIdentical(t.type, self.prog.int_type("int", 4, True))
        self.assertTrue(t.is_complete())

        t = self.prog.array_type(self.prog.int_type("int", 4, True))
        self.assertEqual(t.kind, TypeKind.ARRAY)
        self.assertIsNone(t.primitive)
        self.assertIsNone(t.length)
        self.assertIdentical(t.type, self.prog.int_type("int", 4, True))
        self.assertFalse(t.is_complete())

        self.assertRaises(TypeError, self.prog.array_type, 10, 4)

    def test_function(self):
        t = self.prog.function_type(
            self.prog.void_type(),
            (TypeParameter(self.prog.int_type("int", 4, True), "n"),),
        )
        self.assertEqual(t.kind, TypeKind.FUNCTION)
        self.assertIsNone(t.primitive)
        self.assertEqual(t.language, DEFAULT_LANGUAGE)
        self.assertIdentical(t.type, self.prog.void_type())
        self.assertIdentical(
            t.parameters, (TypeParameter(self.prog.int_type("int", 4, True), "n"),)
        )
        self.assertFalse(t.is_variadic)
        self.assertTrue(t.is_complete())

        self.assertEqual(
            repr(t),
            "prog.function_type(type=prog.void_type(), parameters=(TypeParameter(prog.type('int'), name='n'),), is_variadic=False)",
        )
        self.assertRaises(TypeError, sizeof, t)

        self.assertFalse(
            self.prog.function_type(self.prog.void_type(), (), False).is_variadic
        )
        self.assertTrue(
            self.prog.function_type(self.prog.void_type(), (), True).is_variadic
        )

        self.assertRaisesRegex(
            TypeError, r"must be _drgn\.Type", self.prog.function_type, None, ()
        )
        self.assertRaisesRegex(
            TypeError,
            "must be sequence",
            self.prog.function_type,
            self.prog.void_type(),
            None,
        )
        self.assertRaisesRegex(
            TypeError,
            "must be TypeParameter",
            self.prog.function_type,
            self.prog.void_type(),
            (4,),
        )

    def test_cycle(self):
        t1 = self.prog.struct_type(
            "foo", 8, (TypeMember(lambda: self.prog.pointer_type(t1), "next"),)
        )
        t2 = self.prog.struct_type(
            "foo", 8, (TypeMember(lambda: self.prog.pointer_type(t2), "next"),)
        )
        t3, t4 = (
            self.prog.struct_type(
                "foo", 8, (TypeMember(lambda: self.prog.pointer_type(t4), "next"),)
            ),
            self.prog.struct_type(
                "foo", 8, (TypeMember(lambda: self.prog.pointer_type(t3), "next"),)
            ),
        )
        self.assertIdentical(t1, t2)
        self.assertIdentical(t2, t3)
        self.assertIdentical(t3, t4)

        self.assertEqual(
            repr(t1),
            "prog.struct_type(tag='foo', size=8, members=(TypeMember(prog.type('struct foo *'), name='next', bit_offset=0),))",
        )

    def test_cycle2(self):
        t1 = self.prog.struct_type(
            "list_head",
            16,
            (
                TypeMember(lambda: self.prog.pointer_type(t1), "next"),
                TypeMember(lambda: self.prog.pointer_type(t1), "prev", 8),
            ),
        )
        t2 = self.prog.struct_type(
            "list_head",
            16,
            (
                TypeMember(lambda: self.prog.pointer_type(t2), "next"),
                TypeMember(lambda: self.prog.pointer_type(t2), "prev", 8),
            ),
        )
        self.assertIdentical(t1, t2)

        self.assertEqual(
            repr(t1),
            "prog.struct_type(tag='list_head', size=16, members=(TypeMember(prog.type('struct list_head *'), name='next', bit_offset=0), TypeMember(prog.type('struct list_head *'), name='prev', bit_offset=8)))",
        )

    def test_bad_thunk(self):
        t1 = self.prog.struct_type(
            "foo", 16, (TypeMember(lambda: exec('raise Exception("test")'), "bar"),)
        )
        with self.assertRaisesRegex(Exception, "test"):
            t1.members[0].type
        t1 = self.prog.struct_type("foo", 16, (TypeMember(lambda: 0, "bar"),))
        with self.assertRaisesRegex(
            TypeError, r"TypeMember\(\) callable must return Object or Type"
        ):
            t1.members[0].type

    def test_qualifiers(self):
        self.assertEqual(self.prog.void_type().qualifiers, Qualifiers(0))

        t = self.prog.void_type(qualifiers=Qualifiers.CONST | Qualifiers.VOLATILE)
        self.assertEqual(t.qualifiers, Qualifiers.CONST | Qualifiers.VOLATILE)
        self.assertEqual(
            repr(t),
            f"prog.void_type(qualifiers={repr(Qualifiers.CONST | Qualifiers.VOLATILE)})",
        )

        self.assertIdentical(
            t.qualified(Qualifiers.ATOMIC),
            self.prog.void_type(qualifiers=Qualifiers.ATOMIC),
        )
        self.assertIdentical(t.unqualified(), self.prog.void_type())
        self.assertIdentical(t.qualified(Qualifiers(0)), t.unqualified())

        self.assertRaisesRegex(
            TypeError, "expected Qualifiers", self.prog.void_type, qualifiers=1.5
        )

    def test_unaliased(self):
        self.assertIdentical(
            self.prog.typedef_type(
                "loff_t",
                self.prog.typedef_type(
                    "__kernel_loff_t",
                    self.prog.int_type("long long", 8, True),
                ),
            ).unaliased(),
            self.prog.int_type("long long", 8, True),
        )

    def test_unalias_not_typedef(self):
        self.assertIdentical(
            self.prog.int_type("long long", 8, True).unaliased(),
            self.prog.int_type("long long", 8, True),
        )

    def test_unalias_qualifiers(self):
        self.assertIdentical(
            self.prog.typedef_type(
                "CINT",
                self.prog.int_type("int", 4, True, qualifiers=Qualifiers.CONST),
                qualifiers=Qualifiers.VOLATILE,
            ).unaliased(),
            self.prog.int_type(
                "int", 4, True, qualifiers=Qualifiers.CONST | Qualifiers.VOLATILE
            ),
        )

    def test_unaliased_kind(self):
        self.assertEqual(
            self.prog.typedef_type(
                "loff_t",
                self.prog.typedef_type(
                    "__kernel_loff_t",
                    self.prog.int_type("long long", 8, True),
                ),
            ).unaliased_kind(),
            TypeKind.INT,
        )

    def test_unaliased_kind_not_typedef(self):
        self.assertEqual(
            self.prog.int_type("long long", 8, True).unaliased_kind(),
            TypeKind.INT,
        )

    def test_language(self):
        self.assertEqual(self.prog.void_type(language=None).language, DEFAULT_LANGUAGE)
        self.assertEqual(self.prog.void_type(language=Language.C).language, Language.C)

        self.assertEqual(
            self.prog.int_type("int", 4, True, language=Language.CPP).language,
            Language.CPP,
        )

    def test_language_repr(self):
        self.assertEqual(
            repr(self.prog.void_type(language=Language.CPP)),
            "prog.void_type(language=Language.CPP)",
        )

    def test__repr_pretty_(self):
        t = self.prog.void_type()
        assertReprPrettyEqualsStr(t)

    def test_different_programs_compound(self):
        self.assertRaisesRegex(
            ValueError,
            "object is from different program",
            self.prog.struct_type,
            None,
            4,
            (TypeMember(mock_program().int_type("int", 4, True)),),
        )

    def test_different_programs_compound_callback(self):
        with self.assertRaisesRegex(ValueError, "objects are from different program"):
            self.prog.struct_type(
                None, 4, (TypeMember(lambda: mock_program().int_type("int", 4, True)),)
            ).members[0].type

    def test_different_programs_enum(self):
        self.assertRaisesRegex(
            ValueError,
            "type is from different program",
            self.prog.enum_type,
            None,
            mock_program().int_type("int", 4, True),
            (),
        )

    def test_different_programs_typedef(self):
        self.assertRaisesRegex(
            ValueError,
            "type is from different program",
            self.prog.typedef_type,
            "INT",
            mock_program().int_type("int", 4, True),
        )

    def test_different_programs_pointer(self):
        self.assertRaisesRegex(
            ValueError,
            "type is from different program",
            self.prog.pointer_type,
            mock_program().int_type("int", 4, True),
        )

    def test_different_programs_array(self):
        self.assertRaisesRegex(
            ValueError,
            "type is from different program",
            self.prog.pointer_type,
            mock_program().int_type("int", 4, True),
        )

    def test_different_programs_function_return(self):
        self.assertRaisesRegex(
            ValueError,
            "type is from different program",
            self.prog.function_type,
            mock_program().int_type("int", 4, True),
            (),
        )

    def test_different_programs_function_parameter(self):
        self.assertRaisesRegex(
            ValueError,
            "object is from different program",
            self.prog.function_type,
            self.prog.void_type(),
            (TypeParameter(mock_program().int_type("int", 4, True)),),
        )

    def test_different_programs_function_parameter_callback(self):
        with self.assertRaisesRegex(ValueError, "objects are from different programs"):
            self.prog.function_type(
                self.prog.void_type(),
                (TypeParameter(lambda: mock_program().int_type("int", 4, True)),),
            ).parameters[0].type

    def test_sizeof_default_prog(self):
        self.assertRaises(NoDefaultProgramError, sizeof, "int")
        with with_default_prog(self.prog):
            self.assertEqual(sizeof("int"), 4)
            self.assertRaises(LookupError, sizeof, "foo")

    def test_offsetof_default_prog(self):
        self.types.append(self.point_type)
        self.assertRaises(NoDefaultProgramError, offsetof, "struct point", "y")
        with with_default_prog(self.prog):
            self.assertEqual(offsetof("struct point", "y"), 4)
            self.assertRaises(LookupError, offsetof, "struct foo", "y")


class TestTypeEnumerator(MockProgramTestCase):
    def test_init(self):
        e = TypeEnumerator("a", 1)
        self.assertEqual(e.name, "a")
        self.assertEqual(e.value, 1)

        self.assertRaises(TypeError, TypeEnumerator, "a", None)
        self.assertRaises(TypeError, TypeEnumerator, None, 1)

    def test_repr(self):
        e = TypeEnumerator("a", 1)
        self.assertEqual(repr(e), "TypeEnumerator('a', 1)")

    def test_sequence(self):
        e = TypeEnumerator("a", 1)
        name, value = e
        self.assertEqual(name, "a")
        self.assertEqual(value, 1)
        self.assertEqual(list(e), ["a", 1])

    def test_cmp(self):
        self.assertEqual(TypeEnumerator("a", 1), TypeEnumerator(name="a", value=1))
        self.assertNotEqual(TypeEnumerator("a", 1), TypeEnumerator("a", 2))
        self.assertNotEqual(TypeEnumerator("b", 1), TypeEnumerator("a", 1))


class TestTypeMember(MockProgramTestCase):
    def test_init(self):
        m = TypeMember(self.prog.void_type())
        self.assertIdentical(m.object, Object(self.prog, self.prog.void_type()))
        self.assertIdentical(m.type, self.prog.void_type())
        self.assertIsNone(m.name)
        self.assertEqual(m.bit_offset, 0)
        self.assertEqual(m.offset, 0)
        self.assertIsNone(m.bit_field_size)

        m = TypeMember(Object(self.prog, self.prog.void_type()), "foo")
        self.assertIdentical(m.object, Object(self.prog, self.prog.void_type()))
        self.assertIdentical(m.type, self.prog.void_type())
        self.assertEqual(m.name, "foo")
        self.assertEqual(m.bit_offset, 0)
        self.assertEqual(m.offset, 0)
        self.assertIsNone(m.bit_field_size)

        m = TypeMember(self.prog.void_type(), "foo", 8)
        self.assertIdentical(m.object, Object(self.prog, self.prog.void_type()))
        self.assertIdentical(m.type, self.prog.void_type())
        self.assertEqual(m.name, "foo")
        self.assertEqual(m.bit_offset, 8)
        self.assertEqual(m.offset, 1)
        self.assertIsNone(m.bit_field_size)

        self.assertRaises(TypeError, TypeMember, None)
        self.assertRaises(TypeError, TypeMember, self.prog.void_type(), 1)
        self.assertRaises(TypeError, TypeMember, self.prog.void_type(), "foo", None)

    def test_callable(self):
        m = TypeMember(self.prog.void_type)
        self.assertIdentical(m.object, Object(self.prog, self.prog.void_type()))
        self.assertIdentical(m.type, self.prog.void_type())

        m = TypeMember(lambda: Object(self.prog, self.prog.int_type("int", 4, True)))
        self.assertIdentical(
            m.object, Object(self.prog, self.prog.int_type("int", 4, True))
        )
        self.assertIdentical(m.type, self.prog.int_type("int", 4, True))

        m = TypeMember(lambda: None)
        self.assertRaises(TypeError, getattr, m, "type")

    def test_repr(self):
        m = TypeMember(self.prog.void_type, name="foo")
        self.assertEqual(
            repr(m), "TypeMember(prog.type('void'), name='foo', bit_offset=0)"
        )

        m = TypeMember(self.prog.void_type)
        self.assertEqual(repr(m), "TypeMember(prog.type('void'), bit_offset=0)")

        m = TypeMember(
            Object(self.prog, self.prog.int_type("int", 4, True), bit_field_size=1)
        )
        self.assertEqual(
            repr(m), "TypeMember(Object(prog, 'int', bit_field_size=1), bit_offset=0)"
        )

        m = TypeMember(lambda: None)
        self.assertRaises(TypeError, repr, m)


class TestTypeParameter(MockProgramTestCase):
    def test_init(self):
        p = TypeParameter(self.prog.void_type())
        self.assertIdentical(
            p.default_argument, Object(self.prog, self.prog.void_type())
        )
        self.assertIdentical(p.type, self.prog.void_type())
        self.assertIsNone(p.name)

        p = TypeParameter(
            Object(self.prog, self.prog.int_type("int", 4, True), 5), "foo"
        )
        self.assertIdentical(
            p.default_argument, Object(self.prog, self.prog.int_type("int", 4, True), 5)
        )
        self.assertIdentical(p.type, self.prog.int_type("int", 4, True))
        self.assertEqual(p.name, "foo")

        self.assertRaises(TypeError, TypeParameter, None)
        self.assertRaises(TypeError, TypeParameter, self.prog.void_type(), 1)

    def test_callable(self):
        p = TypeParameter(self.prog.void_type)
        self.assertIdentical(
            p.default_argument, Object(self.prog, self.prog.void_type())
        )
        self.assertIdentical(p.type, self.prog.void_type())

        p = TypeParameter(lambda: Object(self.prog, self.prog.int_type("int", 4, True)))
        self.assertIdentical(
            p.default_argument, Object(self.prog, self.prog.int_type("int", 4, True))
        )
        self.assertIdentical(p.type, self.prog.int_type("int", 4, True))

        p = TypeParameter(lambda: None)
        self.assertRaises(TypeError, getattr, p, "type")

    def test_repr(self):
        p = TypeParameter(self.prog.void_type, name="foo")
        self.assertEqual(repr(p), "TypeParameter(prog.type('void'), name='foo')")

        p = TypeParameter(self.prog.void_type)
        self.assertEqual(repr(p), "TypeParameter(prog.type('void'))")

        p = TypeParameter(lambda: None)
        self.assertRaises(TypeError, repr, p)


class TestTypeTemplateParameter(MockProgramTestCase):
    def test_init(self):
        p = TypeTemplateParameter(self.prog.void_type())
        self.assertIdentical(p.argument, self.prog.void_type())
        self.assertIsNone(p.name)
        self.assertFalse(p.is_default)

        p = TypeTemplateParameter(
            Object(self.prog, self.prog.int_type("int", 4, True), 5), "foo", True
        )
        self.assertIdentical(
            p.argument, Object(self.prog, self.prog.int_type("int", 4, True), 5)
        )
        self.assertEqual(p.name, "foo")
        self.assertTrue(p.is_default)

        self.assertRaises(TypeError, TypeTemplateParameter, None)
        self.assertRaisesRegex(
            ValueError,
            "must not be absent Object",
            TypeTemplateParameter,
            Object(self.prog, "int"),
        )
        self.assertRaises(TypeError, TypeTemplateParameter, self.prog.void_type(), 1)
        self.assertRaises(
            TypeError, TypeTemplateParameter, self.prog.void_type(), None, None
        )

    def test_callable(self):
        p = TypeTemplateParameter(self.prog.void_type)
        self.assertIdentical(p.argument, self.prog.void_type())

        p = TypeTemplateParameter(
            lambda: Object(self.prog, self.prog.int_type("int", 4, True), 5)
        )
        self.assertIdentical(
            p.argument, Object(self.prog, self.prog.int_type("int", 4, True), 5)
        )

        p = TypeTemplateParameter(lambda: None)
        self.assertRaises(TypeError, getattr, p, "argument")
        p = TypeTemplateParameter(lambda: Object(self.prog, "int"))
        self.assertRaisesRegex(
            ValueError, "must not return absent Object", getattr, p, "argument"
        )

    def test_repr(self):
        p = TypeTemplateParameter(self.prog.void_type, name="foo", is_default=True)
        self.assertEqual(
            repr(p),
            "TypeTemplateParameter(prog.type('void'), name='foo', is_default=True)",
        )

        p = TypeTemplateParameter(self.prog.void_type)
        self.assertEqual(repr(p), "TypeTemplateParameter(prog.type('void'))")

        p = TypeTemplateParameter(lambda: None)
        self.assertRaises(TypeError, repr, p)


class TestAlignof(TestCase):
    def setUp(self):
        self.prog = Program(Platform(Architecture.X86_64))

    def test_scalar(self):
        self.assertEqual(alignof(self.prog.int_type("char", 1, True)), 1)
        self.assertEqual(alignof(self.prog.int_type("short", 2, True)), 2)
        self.assertEqual(alignof(self.prog.int_type("int", 4, True)), 4)
        self.assertEqual(alignof(self.prog.int_type("long long", 8, True)), 8)
        self.assertEqual(alignof(self.prog.int_type("__int128", 16, True)), 16)
        self.assertEqual(alignof(self.prog.bool_type("_Bool", 1)), 1)
        self.assertEqual(alignof(self.prog.float_type("float", 4)), 4)
        self.assertEqual(alignof(self.prog.float_type("double", 8)), 8)
        self.assertEqual(alignof(self.prog.float_type("long double", 16)), 16)
        self.assertEqual(alignof(self.prog.pointer_type(self.prog.void_type())), 8)

    def test_empty_struct(self):
        self.assertEqual(alignof(self.prog.struct_type(None, 0, ())), 1)

    def test_struct(self):
        self.assertEqual(
            alignof(
                self.prog.struct_type(
                    None,
                    16,
                    (
                        TypeMember(self.prog.int_type("int", 4, True), "i", 0),
                        TypeMember(self.prog.int_type("long long", 8, True), "ll", 64),
                    ),
                )
            ),
            8,
        )

    def test_union(self):
        self.assertEqual(
            alignof(
                self.prog.struct_type(
                    None,
                    8,
                    (
                        TypeMember(self.prog.int_type("int", 4, True), "i"),
                        TypeMember(self.prog.int_type("long long", 8, True), "ll"),
                    ),
                )
            ),
            8,
        )

    def test_enum(self):
        self.assertEqual(
            alignof(self.prog.enum_type(None, self.prog.int_type("int", 4, True), ())),
            4,
        )

    def test_typedef(self):
        self.assertEqual(
            alignof(self.prog.typedef_type("INT", self.prog.int_type("int", 4, True))),
            4,
        )

    def test_array(self):
        self.assertEqual(
            alignof(self.prog.array_type(self.prog.int_type("int", 4, True), 8)), 4
        )

    def test_incomplete_type(self):
        self.assertRaises(TypeError, alignof, self.prog.void_type())

    def test_cycle(self):
        type = self.prog.struct_type(None, 8, (TypeMember(lambda: type, "foo"),))
        self.assertRaises(RecursionError, alignof, type)

    def test_default_prog(self):
        self.assertRaises(NoDefaultProgramError, alignof, "int")
        with with_default_prog(self.prog):
            self.assertEqual(alignof("int"), 4)
            self.assertRaises(LookupError, alignof, "struct foo")



# Copyright (c) 2024 Oracle and/or its affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later
from drgn.helpers.linux.module import (
    address_to_module,
    find_module,
    for_each_module,
    module_address_regions,
    module_percpu_region,
    module_taints,
)
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


class TestListModules(LinuxKernelTestCase):
    def test_for_each_module(self):
        with open("/proc/modules") as f:
            sys_modules = set(line.split(maxsplit=1)[0] for line in f)
        drgn_modules = set()
        for module in for_each_module(self.prog):
            drgn_modules.add(module.name.string_().decode())

        self.assertEqual(sys_modules, drgn_modules)


@skip_unless_have_test_kmod
class TestModules(LinuxKernelTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.mod = find_module(cls.prog, "drgn_test")

    def test_module_percpu_region(self):
        pcpu_addr = self.prog.symbol("drgn_test_percpu_static").address
        start, size = module_percpu_region(self.mod)
        if start == 0:
            self.skipTest("No module percpu region on !SMP")
        self.assertTrue(start <= pcpu_addr <= start + size)

    def test_module_address_regions(self):
        regions = module_address_regions(self.mod)

        def assertInRegions(addr):
            for start, size in regions:
                if start <= addr < start + size:
                    break
            else:
                self.fail(f"address {addr:x} not found in drgn_test module regions")

            self.assertEqual(address_to_module(self.prog, addr), self.mod)

        # function symbol (should be in .text)
        assertInRegions(self.prog.symbol("drgn_test_function").address)
        # variable symbol (should be in .data)
        assertInRegions(self.prog.symbol("drgn_test_empty_list").address)
        # constant variable (should be in .rodata)
        assertInRegions(self.prog.symbol("drgn_test_have_maple_tree").address)

    def test_module_taints(self):
        self.assertIn("O", module_taints(self.mod))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import logging
import os
from pathlib import Path
import sys
import tempfile
import unittest.mock

import _drgn_util.plugins
from _drgn_util.plugins import call_plugins
from tests import TestCase, modifyenv


class TestPlugins(TestCase):
    def setUp(self):
        # Clear the plugin and hook caches before each test.
        _drgn_util.plugins._plugins = None
        _drgn_util.plugins._hooks.clear()

        # pkg_resources caches distributions on import. Delete it before each
        # test so that it is reloaded.
        sys.modules.pop("pkg_resources", None)

        # These tests change these environment variables and sys.path, so
        # restore them after each test.
        self.enterContext(
            modifyenv({"DRGN_PLUGINS": None, "DRGN_DISABLE_PLUGINS": None})
        )
        self.addCleanup(setattr, sys, "path", list(sys.path))

        # Delete modules imported by each test so that we can reuse the same
        # module names.
        def restore_modules(old_modules):
            for new_module in set(sys.modules) - old_modules:
                sys.modules.pop(new_module, None)

        self.addCleanup(restore_modules, set(sys.modules))

    @staticmethod
    def _create_plugin(dir):
        plugin_path = Path(dir) / "test_plugin.py"
        plugin_path.write_text(
            """\
def drgn_test_hook(call_me):
    call_me()
"""
        )
        return plugin_path

    @staticmethod
    def _create_dist_info(dir, module_name="test_plugin", entry_point_name="test"):
        dist_info_dir = Path(dir) / f"{module_name}-1.0.dist-info"
        dist_info_dir.mkdir()
        (dist_info_dir / "METADATA").write_text(
            f"""\
Metadata-Version: 1.1
Name: {module_name}
Version: 1.0
"""
        )
        (dist_info_dir / "entry_points.txt").write_text(
            f"""\
[drgn.plugins]
{entry_point_name} = {module_name}
"""
        )

    def test_entry_point(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            self._create_plugin(temp_dir)
            self._create_dist_info(temp_dir)
            sys.path.insert(0, temp_dir)

            call_me = unittest.mock.Mock()
            call_plugins("drgn_test_hook", call_me)
            call_me.assert_called_once()

    def test_drgn_disable_plugins_envvar_all(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            self._create_plugin(temp_dir)
            self._create_dist_info(temp_dir)
            sys.path.insert(0, temp_dir)
            os.environ["DRGN_DISABLE_PLUGINS"] = "*"

            call_me = unittest.mock.Mock()
            call_plugins("drgn_test_hook", call_me)
            call_me.assert_not_called()

    def test_drgn_disable_plugins_envvar_specific(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            self._create_plugin(temp_dir)
            self._create_dist_info(temp_dir)
            sys.path.insert(0, temp_dir)
            os.environ["DRGN_DISABLE_PLUGINS"] = "foo,test"

            call_me = unittest.mock.Mock()
            call_plugins("drgn_test_hook", call_me)
            call_me.assert_not_called()

    def test_drgn_plugins_envvar_path(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            plugin_path = self._create_plugin(temp_dir)
            os.environ["DRGN_PLUGINS"] = f"test={plugin_path}"

            call_me = unittest.mock.Mock()
            call_plugins("drgn_test_hook", call_me)
            call_me.assert_called_once()

    def test_drgn_plugins_envvar_module(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            plugin_path = self._create_plugin(temp_dir)
            sys.path.insert(0, temp_dir)
            os.environ["DRGN_PLUGINS"] = f"test={plugin_path.stem}"

            call_me = unittest.mock.Mock()
            call_plugins("drgn_test_hook", call_me)
            call_me.assert_called_once()

    def test_drgn_plugins_envvar_precedence(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            self._create_plugin(temp_dir)
            self._create_dist_info(temp_dir)
            sys.path.insert(0, temp_dir)
            os.environ["DRGN_DISABLE_PLUGINS"] = "*"
            os.environ["DRGN_PLUGINS"] = "test"

            call_me = unittest.mock.Mock()
            call_plugins("drgn_test_hook", call_me)
            call_me.assert_called_once()

    def test_priority(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            (Path(temp_dir) / "test_plugin1.py").write_text(
                """\
def drgn_test_hook(call_me):
    call_me(1)
drgn_test_hook.drgn_priority = 75
"""
            )
            (Path(temp_dir) / "test_plugin2.py").write_text(
                """\
def drgn_test_hook(call_me):
    call_me(2)
drgn_test_hook.drgn_priority = 25
"""
            )
            (Path(temp_dir) / "test_plugin3.py").write_text(
                """\
def drgn_test_hook(call_me):
    call_me(3)
"""
            )
            self._create_dist_info(temp_dir, "test_plugin1", "test1")
            self._create_dist_info(temp_dir, "test_plugin2", "test2")
            self._create_dist_info(temp_dir, "test_plugin3", "test3")
            sys.path.insert(0, temp_dir)

            call_me = unittest.mock.Mock()
            call_plugins("drgn_test_hook", call_me)
            self.assertEqual(
                call_me.call_args_list,
                [unittest.mock.call(2), unittest.mock.call(3), unittest.mock.call(1)],
            )

    def test_plugin_exception(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            (Path(temp_dir) / "test_plugin.py").write_text('raise Exception("foo")\n')
            self._create_dist_info(temp_dir)
            sys.path.insert(0, temp_dir)

            with self.assertLogs(logging.getLogger("drgn.plugins"), "WARNING") as cm:
                call_plugins("drgn_test_hook")
            self.assertTrue(
                any(
                    message.startswith("WARNING:drgn.plugins:failed to load 'test")
                    for message in cm.output
                ),
                msg=f"no match in {cm.output}",
            )

    def test_hook_exception(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            (Path(temp_dir) / "test_plugin.py").write_text(
                """\
def drgn_test_hook():
    raise Exception("foo")
"""
            )
            self._create_dist_info(temp_dir)
            sys.path.insert(0, temp_dir)

            with self.assertLogs(logging.getLogger("drgn.plugins"), "WARNING") as cm:
                call_plugins("drgn_test_hook")
            self.assertTrue(
                any(
                    message.startswith(
                        "WARNING:drgn.plugins:'test' drgn_test_hook failed:"
                    )
                    for message in cm.output
                ),
                msg=f"no match in {cm.output}",
            )

    def test_missing_entry_point(self):
        os.environ["DRGN_PLUGINS"] = "__non__existent__entrypoint__"
        with self.assertLogs(logging.getLogger("drgn.plugins"), "WARNING") as cm:
            call_plugins("drgn_test_hook")
        self.assertTrue(
            any(
                message.startswith(
                    "WARNING:drgn.plugins:not found: '__non__existent__entrypoint__'"
                )
                for message in cm.output
            ),
            msg=f"no match in {cm.output}",
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import os

from drgn import Object
from drgn.helpers.linux.user import find_user, for_each_user, kuid_val
from tests.linux_kernel import LinuxKernelTestCase, fork_and_stop


class TestUser(LinuxKernelTestCase):
    # Try a few UIDs in case the the hash function changes in the future.
    UIDS = frozenset({0, 430, 1000, 65537})

    def test_kuid_val(self):
        self.assertIdentical(kuid_val(self.prog["init_cred"].uid), 0)
        self.assertIdentical(kuid_val(Object(self.prog, "uid_t", 0)), 0)
        self.assertIdentical(kuid_val(0), 0)

    def test_find_user(self):
        for uid in self.UIDS:
            with fork_and_stop(os.setuid, uid):
                found_uid = find_user(self.prog, uid).uid.val.value_()
            self.assertEqual(found_uid, uid)

    def test_for_each_user(self):
        with contextlib.ExitStack() as stack:
            for uid in self.UIDS:
                stack.enter_context(fork_and_stop(os.setuid, uid))
            found_uids = {user.uid.val.value_() for user in for_each_user(self.prog)}
        self.assertTrue(self.UIDS.issubset(found_uids))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
drgn consists of a core C extension and supporting Python code. It also makes
use of type hints. As a result, its documentation generation has a few
requirements:

1. It must work without compiling the C extension, which can't be done on Read
   the Docs because of missing dependencies.
2. It must support generating documentation from type hints (ideally with
   proper markup rather than by including the raw type annotations).
3. It must support type hint stub files.
4. It must support classes/functions/etc. which are defined in one module but
   should canonically be documented in another. This is common for C extensions
   that are wrapped by a higher-level Python module.

The main existing solutions are ruled out by these requirements:

1. sphinx.ext.autodoc (and other solutions based on runtime introspection)
   require excluding the C extension (e.g., with autodoc_mock_imports) and
   providing the documentation for it elsewhere. Additionally, type hints from
   stub files are not available at runtime, so extensions like
   sphinx-autodoc-typehints and sphinx.ext.autodoc.typehints won't work.
2. sphinx.ext.autoapi doesn't generate markup for type hints and doesn't have
   any support for objects which should documented under a different name than
   they were defined. It also only supports documenting directory trees, not
   individual files.

This extension addresses these requirements. In the future, it may be
worthwhile to make it a standalone package, as I imagine other projects that
make heavy use of C extensions have encountered similar issues.

Overall, it works by parsing Python source code and stub files (drgndoc.parse),
building a tree representing the namespace (drgndoc.namespace), and using that
namespace to resolve definitions and type annotations to generate markup
(drgndoc.format).

This also provides a script that can generate docstring definitions from a stub
file for the C extension itself (drgndoc.docstrings).
"""

import ast
import os.path
import re
from typing import Any, Dict, Optional, cast

import docutils.nodes
import docutils.parsers.rst.directives
import docutils.statemachine
import sphinx.addnodes
import sphinx.application
import sphinx.environment
import sphinx.util.docutils
import sphinx.util.logging
import sphinx.util.nodes

from drgndoc.commands import CommandFormatter
from drgndoc.format import Formatter
from drgndoc.namespace import Namespace, ResolvedNode
from drgndoc.parse import (
    Class,
    DocumentedNode,
    Import,
    ImportFrom,
    Module,
    Node,
    Variable,
    parse_paths,
)
from drgndoc.util import dot_join

logger = sphinx.util.logging.getLogger(__name__)


# Needed for type checking.
class DrgnDocBuildEnvironment(sphinx.environment.BuildEnvironment):
    drgndoc_namespace: Namespace
    drgndoc_formatter: Formatter
    drgndoc_command_formatter: CommandFormatter


def drgndoc_init(app: sphinx.application.Sphinx) -> None:
    env = cast(DrgnDocBuildEnvironment, app.env)

    paths = [os.path.join(app.confdir, path) for path in app.config.drgndoc_paths]
    env.drgndoc_namespace = Namespace(parse_paths(paths, logger.warning))
    env.drgndoc_formatter = Formatter(
        env.drgndoc_namespace,
        [
            (re.compile(pattern), repl)
            for pattern, repl in app.config.drgndoc_substitutions
        ],
    )
    env.drgndoc_command_formatter = CommandFormatter(env.drgndoc_namespace)


# Sphinx looks up type annotations as py:class references. This doesn't work
# for type aliases, which are py:data. See
# https://github.com/sphinx-doc/sphinx/issues/10785. This hack intercepts
# missing py:class references, and if they resolve to a variable annotated as
# TypeAlias, retries them as py:data.
def missing_reference(
    app: sphinx.application.Sphinx,
    env: DrgnDocBuildEnvironment,
    node: sphinx.addnodes.pending_xref,
    contnode: docutils.nodes.Element,
) -> Optional[docutils.nodes.Element]:
    if node.get("refdomain") == "py":
        reftarget = node.get("reftarget")
        if reftarget and node.get("reftype") == "class":
            resolved = env.drgndoc_namespace.resolve_global_name(reftarget)
            if not isinstance(resolved, ResolvedNode):
                py_module = node.get("py:module", "")
                if py_module:
                    resolved = env.drgndoc_namespace.resolve_global_name(
                        dot_join(py_module, reftarget)
                    )
                classes = node.get("classes")
                if not isinstance(resolved, ResolvedNode) and classes:
                    resolved = env.drgndoc_namespace.resolve_global_name(
                        dot_join(py_module, *classes, reftarget)
                    )
            if (
                isinstance(resolved, ResolvedNode)
                and isinstance(resolved.node, Variable)
                and isinstance(resolved.node.annotation, ast.Name)
                and resolved.node.annotation.id == "TypeAlias"
            ):
                node.attributes["reftype"] = "data"
                return env.domains["py"].resolve_xref(
                    env,
                    node.get("refdoc"),
                    app.builder,
                    "data",
                    reftarget,
                    node,
                    contnode,
                )
    return None


class DrgnDocDirective(sphinx.util.docutils.SphinxDirective):
    env: DrgnDocBuildEnvironment

    required_arguments = 1
    optional_arguments = 0
    option_spec = {
        "exclude": docutils.parsers.rst.directives.unchanged,
    }

    def run(self) -> Any:
        parts = []
        py_module = self.env.ref_context.get("py:module")
        if py_module:
            parts.append(py_module)
        py_classes = self.env.ref_context.get("py:classes", [])
        if py_classes:
            parts.extend(py_classes)
        parts.append(self.arguments[0])
        name = ".".join(parts)
        resolved = self.env.drgndoc_namespace.resolve_global_name(name)
        if not isinstance(resolved, ResolvedNode):
            logger.warning("name %r not found", resolved)
            return []
        if not resolved.node.has_docstring():
            logger.warning("name %r is not documented", resolved.qualified_name())
            return []

        docnode = docutils.nodes.section()
        self._run(name, "", self.arguments[0], resolved, docnode)
        return docnode.children

    def _run(
        self,
        top_name: str,
        attr_name: str,
        name: str,
        resolved: ResolvedNode[Node],
        docnode: docutils.nodes.Node,
    ) -> None:
        exclude_pattern = self.options.get("exclude")
        if exclude_pattern is not None and re.fullmatch(exclude_pattern, attr_name):
            return

        if isinstance(resolved.node, (Import, ImportFrom)):
            # Only include imports that are explicitly aliased (i.e., import
            # ... as ... or from ... import ... as ...).
            # TODO: we should also include imports listed in __all__.
            if not resolved.node.aliased:
                return
            imported = self.env.drgndoc_namespace.resolve_name_in_scope(
                resolved.modules, resolved.classes, resolved.name
            )
            if not isinstance(imported, ResolvedNode):
                return
            resolved = imported

        resolved = cast(ResolvedNode[DocumentedNode], resolved)

        if isinstance(resolved.node, Module):
            return self._run_module(
                top_name, attr_name, cast(ResolvedNode[Module], resolved), docnode
            )

        lines = self.env.drgndoc_formatter.format(
            resolved,
            name,
            self.env.ref_context.get("py:module", ""),
            ".".join(self.env.ref_context.get("py:classes", ())),
        )
        if not lines:
            # Not documented. Ignore it.
            return

        sourcename = ""
        if resolved.modules and resolved.modules[-1].node.path:
            sourcename = resolved.modules[-1].node.path
        if sourcename:
            self.env.note_dependency(sourcename)
        contents = docutils.statemachine.StringList(lines, sourcename)
        contents.append("", sourcename)

        self.state.nested_parse(contents, 0, docnode)
        if isinstance(resolved.node, Class):
            for desc in reversed(docnode.children):
                if isinstance(desc, sphinx.addnodes.desc):
                    break
            else:
                logger.warning("desc node not found")
                return
            for desc_content in reversed(desc.children):
                if isinstance(desc_content, sphinx.addnodes.desc_content):
                    break
            else:
                logger.warning("desc_content node not found")
                return

            py_classes = self.env.ref_context.setdefault("py:classes", [])
            py_classes.append(resolved.name)
            self.env.ref_context["py:class"] = resolved.name
            for member in resolved.attrs():
                if member.name != "__init__":
                    self._run(
                        top_name,
                        dot_join(attr_name, member.name),
                        member.name,
                        member,
                        desc_content,
                    )
            py_classes.pop()
            self.env.ref_context["py:class"] = py_classes[-1] if py_classes else None

    def _run_module(
        self,
        top_name: str,
        attr_name: str,
        resolved: ResolvedNode[Module],
        docnode: docutils.nodes.Node,
    ) -> None:
        node = resolved.node
        if node.docstring is None:
            # Not documented. Ignore it.
            return

        try:
            old_py_module = self.env.ref_context["py:module"]
            have_old_py_module = True
        except KeyError:
            have_old_py_module = False

        module_name = dot_join(top_name, attr_name)

        sourcename = node.path or ""
        if sourcename:
            self.env.note_dependency(sourcename)
        contents = docutils.statemachine.StringList(
            [
                ".. py:module:: " + module_name,
                "",
                *node.docstring.splitlines(),
            ],
            sourcename,
        )

        sphinx.util.nodes.nested_parse_with_titles(self.state, contents, docnode)

        # If the module docstring defines any sections, then the contents
        # should go inside of the last one.
        section = docnode
        for child in reversed(docnode.children):
            if isinstance(child, docutils.nodes.section):
                section = child
                break

        attrs = []
        submodules = []
        for attr in resolved.attrs():
            if isinstance(attr.node, Module):
                submodules.append(attr)
            else:
                attrs.append(attr)

        # Submodules are initially sorted by name (guaranteed by
        # parse_package()). Apply any sorting configuration.
        for module_pattern, sort_key_patterns in self.config.drgndoc_submodule_sort:
            if re.fullmatch(module_pattern, module_name):
                # list.sort() is stable, so this preserves the previous order
                # for submodules with the same key.
                def sort_key(attr: ResolvedNode[Node]) -> Any:
                    for pattern, key in sort_key_patterns:
                        if re.fullmatch(pattern, attr.name):
                            return key
                    return 0

                submodules.sort(key=sort_key)

        # Normal attributes go before submodules.
        attrs.extend(submodules)

        for attr in attrs:
            self._run(
                top_name, dot_join(attr_name, attr.name), attr.name, attr, section
            )
        if have_old_py_module:
            self.env.ref_context["py:module"] = old_py_module
        else:
            del self.env.ref_context["py:module"]


class DrgnCommandDirective(sphinx.util.docutils.SphinxDirective):
    env: DrgnDocBuildEnvironment

    required_arguments = 1
    optional_arguments = 0
    has_content = True

    def run(self) -> Any:
        before, sep, after = self.arguments[0].partition(".")
        if sep:
            namespace_name = before
            command_name = after
        else:
            namespace_name = ""
            command_name = before

        if namespace_name:
            name = f"{namespace_name}.{command_name}"
        elif "." in command_name:
            name = f".{command_name}"
        else:
            name = command_name

        self.env.ref_context["std:program"] = name

        # parse_content_to_nodes() was added in Sphinx 7.4. Fall back to an
        # equivalent on older versions.
        if hasattr(self, "parse_content_to_nodes"):
            nodes = self.parse_content_to_nodes(allow_section_headings=True)
        else:
            node = docutils.nodes.Element()
            node.document = self.state.document
            sphinx.util.nodes.nested_parse_with_titles(self.state, self.content, node)
            nodes = node.children

        if nodes:
            target = cast(docutils.nodes.Element, nodes[0])
            std = self.env.get_domain("std")
            std.note_object(  # type: ignore[attr-defined]
                "drgncommand", name, target["ids"][0], location=target
            )

        return nodes


class DrgnDocCommandDirective(sphinx.util.docutils.SphinxDirective):
    env: DrgnDocBuildEnvironment

    required_arguments = 1
    optional_arguments = 0

    def run(self) -> Any:
        before, sep, after = self.arguments[0].partition(".")
        if sep:
            namespace_name = before
            command_name = after
        else:
            namespace_name = ""
            command_name = before

        try:
            namespace = self.env.drgndoc_command_formatter.command_namespaces[
                namespace_name
            ]
        except KeyError:
            logger.warning("drgn command namespace %r not found", namespace_name)
            return []

        try:
            command = namespace[command_name]
        except KeyError:
            if namespace_name:
                logger.warning(
                    "drgn command %r not found in namespace %r",
                    command_name,
                    namespace_name,
                )
            else:
                logger.warning("drgn command %r not found", command_name)
            return []

        docnode = docutils.nodes.section()
        lines = self.env.drgndoc_command_formatter.format(command)
        if not lines:
            return []

        lines = [
            f".. drgncommand:: {self.arguments[0]}",
            "",
            *("    " + line if line else line for line in lines),
        ]

        sourcename = ""
        if command.func.modules and command.func.modules[-1].node.path:
            sourcename = command.func.modules[-1].node.path
        if sourcename:
            self.env.note_dependency(sourcename)
        contents = docutils.statemachine.StringList(lines, sourcename)
        contents.append("", sourcename)
        sphinx.util.nodes.nested_parse_with_titles(self.state, contents, docnode)
        return docnode.children


class DrgnDocCommandNamespaceDirective(sphinx.util.docutils.SphinxDirective):
    env: DrgnDocBuildEnvironment

    required_arguments = 0
    optional_arguments = 1
    option_spec = {
        "enabled": docutils.parsers.rst.directives.unchanged,
    }

    def run(self) -> Any:
        namespace_name = self.arguments[0] if self.arguments else ""
        try:
            namespace = self.env.drgndoc_command_formatter.command_namespaces[
                namespace_name
            ]
        except KeyError:
            logger.warning("drgn command namespace %r not found", namespace_name)
            return []

        enabled = re.compile(self.options.get("enabled", ""))

        command_names = [
            name
            for name, command in namespace.items()
            if enabled.fullmatch(command.enabled)
        ]
        command_names.sort()

        lines = []
        for i, command_name in enumerate(command_names):
            if i != 0:
                lines.append("----")
                lines.append("")

            if namespace_name:
                name = f"{namespace_name}.{command_name}"
            elif "." in command_name:
                name = f".{command_name}"
            else:
                name = command_name
            lines.append(f".. drgndoc-command:: {name}")
            lines.append("")

        docnode = docutils.nodes.section()
        if lines:
            contents = docutils.statemachine.StringList(lines, "")
            contents.append("", "")
            sphinx.util.nodes.nested_parse_with_titles(self.state, contents, docnode)
        return docnode.children


def setup(app: sphinx.application.Sphinx) -> Dict[str, Any]:
    app.connect("builder-inited", drgndoc_init)
    app.connect("missing-reference", missing_reference)
    # List of modules or packages.
    app.add_config_value("drgndoc_paths", [], "env")
    # List of (regex pattern, substitution) to apply to resolved names.
    app.add_config_value("drgndoc_substitutions", [], "env")
    # List of (parent regex pattern, list of (submodule regex pattern, key))
    # controlling sort order of submodules.
    #
    # Submodules are initially sorted by name. For each parent regex pattern
    # matching the fully qualified name of the parent module, the list of
    # submodules is sorted. The sort key is given by the first submodule regex
    # pattern matching the relative name of the subvolume, or 0 if no patterns
    # match.
    app.add_config_value("drgndoc_submodule_sort", [], "env")
    app.add_directive("drgndoc", DrgnDocDirective)
    # Create a drgncommand object type...
    app.add_object_type("drgncommand", "drgncommand")
    # ... but override the directive with our own.
    app.add_directive_to_domain(
        "std", "drgncommand", DrgnCommandDirective, override=True
    )
    app.add_directive("drgndoc-command", DrgnDocCommandDirective)
    app.add_directive("drgndoc-command-namespace", DrgnDocCommandNamespaceDirective)
    return {"env_version": 1, "parallel_read_safe": True, "parallel_write_safe": True}



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from pathlib import Path
import re
import tempfile

from drgn import Object
from drgn.commands._crash.common import CRASH_COMMAND_NAMESPACE
from tests.linux_kernel import possible_cpus, skip_unless_have_test_kmod
from tests.linux_kernel.crash_commands import CrashCommandTestCase


class TestEval(CrashCommandTestCase):
    def test_value(self):
        cmd = self.check_crash_command("eval 10")
        self.assertRegex(cmd.stdout, r"\bhexadecimal: a\b")
        self.assertRegex(cmd.stdout, r"\bdecimal: 10\b")
        self.assertRegex(cmd.stdout, r"\boctal: 12\b")
        self.assertRegex(cmd.stdout, r"\bbinary: 0*1010\b")

        self.assertIdentical(
            cmd.drgn_option.globals["value"], Object(self.prog, "unsigned long", 10)
        )
        self.assertIdentical(
            cmd.drgn_option.globals["signed"], Object(self.prog, "long", 10)
        )
        self.assertIsInstance(cmd.drgn_option.globals["in_units"], str)
        self.assertEqual(cmd.drgn_option.globals["in_hex"], "0xa")
        self.assertEqual(cmd.drgn_option.globals["in_octal"], "0o12")
        self.assertEqual(cmd.drgn_option.globals["in_binary"], "0b1010")

    def test_negate(self):
        cmd = self.check_crash_command("eval -99")
        self.assertRegex(cmd.stdout, r"\bhexadecimal: f+9d\b")
        self.assertRegex(cmd.stdout, r"\bdecimal: [0-9]+  \(-99\)")

        self.assertIdentical(
            cmd.drgn_option.globals["value"], Object(self.prog, "unsigned long", -99)
        )
        self.assertIdentical(
            cmd.drgn_option.globals["signed"], Object(self.prog, "long", -99)
        )

    def test_invert(self):
        cmd = self.check_crash_command("eval ~99")
        self.assertRegex(cmd.stdout, r"\bhexadecimal: f+9c\b")
        self.assertRegex(cmd.stdout, r"\bdecimal: [0-9]+  \(-100\)")

        self.assertIdentical(
            cmd.drgn_option.globals["value"], Object(self.prog, "unsigned long", ~99)
        )
        self.assertIdentical(
            cmd.drgn_option.globals["signed"], Object(self.prog, "long", ~99)
        )

    def test_multiple_unary(self):
        cmd = self.check_crash_command("eval -~--99")
        self.assertRegex(cmd.stdout, r"\bhexadecimal: 64\b")
        self.assertRegex(cmd.stdout, r"\bdecimal: 100\b")

        self.assertIdentical(
            cmd.drgn_option.globals["value"], Object(self.prog, "unsigned long", 100)
        )
        self.assertIdentical(
            cmd.drgn_option.globals["signed"], Object(self.prog, "long", 100)
        )

    def test_units(self):
        for value, in_units in (
            (1024, "1K"),
            (1024 * 1024, "1M"),
            (1024 * 1024 * 1024, "1G"),
        ):
            for expr in (str(value), in_units, in_units.lower()):
                with self.subTest(expr=expr):
                    cmd = self.check_crash_command("eval " + expr)
                    self.assertRegex(
                        cmd.stdout, rf"\bhexadecimal: {value:x}  \({in_units}B\)"
                    )
                    self.assertEqual(cmd.drgn_option.globals["in_units"], in_units)

    def test_negate_with_unit(self):
        cmd = self.check_crash_command("eval -1k")
        self.assertRegex(cmd.stdout, r"\bdecimal: [0-9]+  \(-1024\)")
        self.assertIdentical(
            cmd.drgn_option.globals["value"], Object(self.prog, "unsigned long", -1024)
        )

    def test_bits_set(self):
        cmd = self.check_crash_command("eval -b 10")
        self.assertRegex(cmd.stdout, r"\bbits set: 3 1\n")
        self.assertIdentical(cmd.drgn_option.globals["bits_set"], [1, 3])

    def test_long_long(self):
        cmd = self.check_crash_command("eval -l 10")
        self.assertRegex(
            cmd.stdout,
            r"\bbinary: 0000000000000000000000000000000000000000000000000000000000001010\b",
        )
        self.assertIdentical(
            cmd.drgn_option.globals["value"],
            Object(self.prog, "unsigned long long", 10),
        )

    def test_bits_set_and_long_long(self):
        cmd = self.check_crash_command("eval -bl 10")
        self.assertRegex(
            cmd.stdout,
            r"\bbinary: 0000000000000000000000000000000000000000000000000000000000001010\b",
        )
        self.assertRegex(cmd.stdout, r"\bbits set: 3 1\n")
        self.assertIdentical(cmd.drgn_option.globals["bits_set"], [1, 3])
        self.assertIdentical(
            cmd.drgn_option.globals["value"],
            Object(self.prog, "unsigned long long", 10),
        )

    def test_negative_b(self):
        cmd = self.check_crash_command("eval -b")
        self.assertRegex(cmd.stdout, r"\bdecimal: [0-9]+  \(-11\)")
        self.assertNotIn("bits set", cmd.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["value"], Object(self.prog, "unsigned long", -0xB)
        )

    def test_bits_set_in_negative_b(self):
        cmd = self.check_crash_command("eval -b -b")
        self.assertRegex(cmd.stdout, r"\bdecimal: [0-9]+  \(-11\)")
        self.assertIn("bits set", cmd.stdout)
        self.assertIn("bits_set", cmd.drgn_option.globals)

    def test_binary(self):
        for expr, result in (
            ("3 + 4", 7),
            ("4 - 3", 1),
            ("12 & 5", 4),
            ("(12 | 5)", 13),
            ("12 ^ 5", 9),
            ("12 * 5", 60),
            ("12 % 5", 2),
            ("16 / 5", 3),
            ("(3 << 5)", 96),
            ("(100 >> 1)", 50),
        ):
            cmd = self.check_crash_command("eval " + expr)
            self.assertRegex(cmd.stdout, rf"\bdecimal: {result}\b")
            self.assertIdentical(
                cmd.drgn_option.globals["value"],
                Object(self.prog, "unsigned long", result),
            )

    def test_pipe(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir) / "file"
            CRASH_COMMAND_NAMESPACE.run(self.prog, f"eval (12 | 5) | grep 13 > {path}")
            self.assertRegex(path.read_text(), r"\bdecimal: 13\b")

    def test_redirect(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = Path(temp_dir) / "file"
            CRASH_COMMAND_NAMESPACE.run(self.prog, f"eval (3 << 5) >> {path}")
            self.assertRegex(path.read_text(), r"\bdecimal: 96\b")


class TestP(CrashCommandTestCase):
    def test_object(self):
        cmd = self.check_crash_command("p jiffies")
        self.assertRegex(cmd.stdout, r"jiffies = \([^)]+\)[0-9]+")
        self.assertIdentical(cmd.drgn_option.globals["object"], self.prog["jiffies"])

    @skip_unless_have_test_kmod
    def test_cpuspec(self):
        cmd = self.check_crash_command("p drgn_test_percpu_structs:a")
        matches = re.findall(
            r"^per_cpu\(drgn_test_percpu_structs, ([0-9]+)\) = \(struct drgn_test_percpu_struct\)\{",
            cmd.stdout,
            flags=re.MULTILINE,
        )
        cpus = sorted(possible_cpus())
        self.assertEqual([int(match) for match in matches], cpus)
        self.assertIn("per_cpu(", cmd.drgn_option.stdout)
        self.assertEqual(cmd.drgn_option.globals["object"].cpu, max(cpus))

    def test_member(self):
        cmd = self.check_crash_command("p init_task.pid")
        self.assertRegex(cmd.stdout, r"init_task\.pid = \([^)]*\)0")
        self.assertIn(".pid", cmd.drgn_option.stdout)
        self.assertIdentical(
            cmd.drgn_option.globals["object"], self.prog["init_task"].pid
        )

    @skip_unless_have_test_kmod
    def test_radix(self):
        self.addCleanup(self.prog.config.pop, "crash_radix", None)

        self.run_crash_command("set radix 16")
        cmd = self.run_crash_command("p drgn_test_singular_list_entry")
        self.assertRegex(cmd.stdout, r"value = \([^)]+\)0x0")

        cmd = self.run_crash_command("p -d drgn_test_singular_list_entry")
        self.assertRegex(cmd.stdout, r"value = \([^)]+\)0\b")

        self.run_crash_command("set radix 10")
        cmd = self.run_crash_command("p drgn_test_singular_list_entry")
        self.assertRegex(cmd.stdout, r"value = \([^)]+\)0\b")

        cmd = self.run_crash_command("p -x drgn_test_singular_list_entry")
        self.assertRegex(cmd.stdout, r"value = \([^)]+\)0x0")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import ctypes
import mmap
import os
import os.path
from pathlib import Path
import socket
import stat
import tempfile

from drgn.helpers.linux.fs import (
    d_path,
    decode_file_type,
    dentry_path,
    fget,
    for_each_file,
    for_each_mount,
    inode_for_each_page,
    inode_path,
    inode_paths,
    mount_dst,
    path_lookup,
)
from drgn.helpers.linux.pid import find_task
from tests.linux_kernel import MS_BIND, LinuxKernelTestCase, mlock, mount, umount


class TestFs(LinuxKernelTestCase):
    def test_path_lookup(self):
        with tempfile.NamedTemporaryFile(prefix="drgn-tests-") as f:
            path = path_lookup(self.prog, os.path.abspath(f.name))
            self.assertEqual(
                path.dentry.d_name.name.string_(), os.fsencode(os.path.basename(f.name))
            )

    def test_path_lookup_bind_mount(self):
        with tempfile.NamedTemporaryFile(prefix="drgn-tests-") as f:
            old_mnt = path_lookup(self.prog, os.path.abspath(f.name)).mnt
            mount(f.name, f.name, "", MS_BIND)
            try:
                new_mnt = path_lookup(self.prog, os.path.abspath(f.name)).mnt
                self.assertNotEqual(old_mnt, new_mnt)
            finally:
                umount(f.name)

    def test_d_path(self):
        task = find_task(self.prog, os.getpid())
        self.assertEqual(d_path(task.fs.pwd.address_of_()), os.fsencode(os.getcwd()))

    def test_d_path_dentry_only(self):
        # Since d_path(dentry) picks an arbitrary mount containing the dentry,
        # this should be a directory that is unlikely to be bind mounted
        # anywhere else.
        with tempfile.NamedTemporaryFile(dir="/dev/shm") as f:
            dentry = fget(find_task(self.prog, os.getpid()), f.fileno()).f_path.dentry
            self.assertEqual(d_path(dentry), os.fsencode(f.name))

    def test_d_path_no_internal_mount(self):
        if not os.path.isdir("/sys/kernel/tracing"):
            self.skipTest("The /sys/kernel/tracing directory is not mounted")
        path = path_lookup(self.prog, "/sys/kernel/tracing/trace_pipe")
        # The first mount for this super block is usually MNT_INTERNAL, but we
        # don't want that one. Ensure we skip it.
        self.assertEqual(d_path(path.dentry), b"/sys/kernel/tracing/trace_pipe")

    def test_dentry_path(self):
        pwd = os.fsencode(os.getcwd())
        task = find_task(self.prog, os.getpid())
        self.assertTrue(pwd.endswith(dentry_path(task.fs.pwd.dentry)))

    def test_inode_paths(self):
        with tempfile.TemporaryDirectory(prefix="drgn-tests-") as dir:
            path1 = os.fsencode(os.path.abspath(os.path.join(dir, "a")))
            path2 = os.fsencode(os.path.abspath(os.path.join(dir, "b")))
            with open(path1, "w"):
                os.link(path1, path2)
                with open(path2, "r"):
                    inode = path_lookup(self.prog, path1).dentry.d_inode
                    paths = list(inode_paths(inode))
                    self.assertEqual(len(paths), 2)
                    self.assertTrue(
                        (path1.endswith(paths[0]) and path2.endswith(paths[1]))
                        or (path1.endswith(paths[1]) and path2.endswith(paths[0]))
                    )
                    self.assertIn(inode_path(inode), paths)

    def test_inode_for_each_page(self):
        with tempfile.TemporaryFile(dir="/dev/shm") as f:
            f.write(bytes(2 * mmap.PAGESIZE))
            f.flush()
            inode = fget(find_task(self.prog, os.getpid()), f.fileno()).f_inode.read_()
            with mmap.mmap(f.fileno(), 2 * mmap.PAGESIZE) as map:
                f.close()
                address = ctypes.addressof(ctypes.c_char.from_buffer(map))
                # Make sure the pages are faulted in and stay that way.
                mlock(address, 2 * mmap.PAGESIZE)

                indices = []
                i_mapping = inode.i_mapping.read_()
                for index, page in inode_for_each_page(inode):
                    indices.append(index)
                    self.assertEqual(page.mapping, i_mapping)
                self.assertEqual(indices, [0, 1])

    def test_for_each_mount(self):
        with open("/proc/self/mounts", "rb") as f:
            self.assertEqual(
                {line.split()[1].decode("unicode-escape") for line in f},
                {os.fsdecode(mount_dst(mount)) for mount in for_each_mount(self.prog)},
            )

    def test_for_each_mount_cursor(self):
        fd = os.open("/proc/self/mounts", os.O_RDONLY)
        try:
            # Read a small amount of data so that we leave a cursor in the
            # middle of the mount list, on kernel versions where this happens.
            # Cursors were introduced in v5.8 with 9f6c61f96f2d9 ("proc/mounts:
            # add cursor"), and were eliminated in v6.8 with 2eea9ce4310d8
            # ("mounts: keep list of mounts in an rbtree"). They were marked
            # with the flag MNT_CURSOR, but that was only defined as a
            # preprocessor constant. It's easiest to detect cursors via their
            # NULL superblock, which should never be present.
            os.read(fd, 64)
            for mnt in for_each_mount(self.prog):
                self.assertNotEqual(mnt.mnt.mnt_sb.value_(), 0)
        finally:
            os.close(fd)

    def test_fget(self):
        with tempfile.NamedTemporaryFile(prefix="drgn-tests-") as f:
            file = fget(find_task(self.prog, os.getpid()), f.fileno())
            self.assertEqual(d_path(file.f_path), os.fsencode(os.path.abspath(f.name)))

    def test_for_each_file(self):
        task = find_task(self.prog, os.getpid())
        with os.scandir("/proc/self/fd") as dir:
            # NB: The call to for_each_file() comes first so that it will
            # include the scandir file descriptor.
            self.assertEqual(
                {fd for fd, file in for_each_file(task)},
                {int(entry.name) for entry in dir},
            )

    def test_decode_file_type(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_dir = Path(tmp_dir)

            with socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM) as sock:
                sock.bind(bytes(tmp_dir / "sock"))
                self.assertEqual(
                    decode_file_type((tmp_dir / "sock").stat().st_mode), "SOCK"
                )

            (tmp_dir / "lnk").symlink_to("/dev/null")
            self.assertEqual(
                decode_file_type(
                    os.stat(tmp_dir / "lnk", follow_symlinks=False).st_mode
                ),
                "LNK",
            )

            (tmp_dir / "reg").touch()
            self.assertEqual(decode_file_type((tmp_dir / "reg").stat().st_mode), "REG")

            os.mknod(tmp_dir / "blk", stat.S_IFBLK | 0o600, device=os.makedev(0, 0))
            self.assertEqual(decode_file_type((tmp_dir / "blk").stat().st_mode), "BLK")

            self.assertEqual(decode_file_type(tmp_dir.stat().st_mode), "DIR")

            self.assertEqual(decode_file_type(os.stat("/dev/null").st_mode), "CHR")

            os.mkfifo(tmp_dir / "fifo")
            self.assertEqual(
                decode_file_type((tmp_dir / "fifo").stat().st_mode), "FIFO"
            )

    def test_decode_file_type_unknown(self):
        self.assertEqual(decode_file_type(0), "000000")
        self.assertEqual(decode_file_type(0o030000), "030000")
        self.assertEqual(decode_file_type(0o050000), "050000")
        self.assertEqual(decode_file_type(0o070000), "070000")
        self.assertEqual(decode_file_type(0o110000), "110000")
        self.assertEqual(decode_file_type(0o130000), "130000")
        self.assertEqual(decode_file_type(0o150000), "150000")
        self.assertEqual(decode_file_type(0o170000), "170000")
